"""Custom exceptions."""

from NeuroRuler.utils.constants import (
    deprecated,
    ROTATION_MAX,
    ROTATION_MIN,
    JSON_GUI_CONFIG_PATH,
)


class ComputeCircumferenceOfInvalidSlice(Exception):
    """User attempted to compute circumference of an invalid slice.

    We detect this by noticing that the number of contours in the slice >= global_vars.NUM_CONTOURS_IN_INVALID_SLICE
    AFTER processing (threshold, select largest component, etc.) in NeuroRuler.utils.imgproc.contour().

    Most valid brain slices have only 2 or 3 detectable contours.

    Change the number in global_vars.py, then run ``pytest`` and examine slices given by settings in
    tests/noise_vals.txt. Some valid slices have 6 or 7 contours.

    See NIFTI file (0, 0, 0, 151) for a valid slice with 9 contours. 9 seems like a good limit.
    """

    def __init__(self, num_contours):
        self.message = (
            f"You attempted to compute the circumference of something that isn't a valid brain slice. After processing, contours detected was {num_contours}\n"
            f"Likely just noise or otherwise not a valid brain slice."
        )
        super().__init__(self.message)


class ArraysDifferentShape(Exception):
    def __init__(self):
        self.message = f"Ran into two arrays of different shape when it was necessary that they be of the same shape."
        super().__init__(self.message)


class InvalidColor(Exception):
    def __init__(self, color: str):
        self.message = f"Invalid color {color} specified."
        super().__init__(self.message)


class RotationOutOfBounds(Exception):
    """Should never be encountered in the GUI because our sliders have min and max.
    This will be for CLI arguments."""

    def __init__(self, theta: int, axis: str):
        self.message = f"{axis} rotation value {theta} out of bounds. Min rotation value is {ROTATION_MIN} degrees, and max is {ROTATION_MAX}."
        super().__init__(self.message)


class InvalidJSONField(Exception):
    def __init__(self, field: str, expected: str):
        """``field`` is the name of the invalid field

        ``expected`` is some information about what the expected value should be"""
        self.message = f"The field {field} in {JSON_GUI_CONFIG_PATH} has an invalid type or value.\nExpected: {expected}"
        super().__init__(self.message)



"""Parse config JSON and CLI arguments to set global settings in ``gui_settings.py`` and ``cli_settings.py``."""

# TODO: (Eric) I think most of this should be refactored into the CLI/GUI packages respectively,
# and we should just keep the general parse functions here

import argparse
import json
from pathlib import Path
import string

import NeuroRuler.utils.cli_settings as cli_settings
import NeuroRuler.utils.gui_settings as gui_settings
import NeuroRuler.utils.constants as constants
import NeuroRuler.utils.exceptions as exceptions

JSON_SETTINGS: dict = dict()
"""Dict of settings resulting from JSON file parsing. Global within this file."""


def parse_cli() -> None:
    """Parse CLI (non-GUI) args and set settings in ``cli_settings.py``.

    :return: None"""
    parser = argparse.ArgumentParser(
        prog="NeuroRuler",
        description="A program that calculates head circumference from MRI data (``.nii``, ``.nii.gz``, ``.nrrd``).",
    )

    parser.add_argument("-d", "--debug", help="print debug info", action="store_true")
    parser.add_argument(
        "-r", "--raw", help='print just the "raw" circumference', action="store_true"
    )
    parser.add_argument("-x", "--x", type=int, help="x rotation (in degrees)")
    parser.add_argument("-y", "--y", type=int, help="y rotation (in degrees)")
    parser.add_argument("-z", "--z", type=int, help="z rotation (in degrees)")
    parser.add_argument("-s", "--slice", type=int, help="slice (Z slice, 0-indexed)")
    parser.add_argument(
        "-c", "--conductance", type=float, help="conductance smoothing parameter"
    )
    parser.add_argument("-i", "--iterations", type=int, help="smoothing iterations")
    parser.add_argument(
        "-t", "--step", type=float, help="time step (smoothing parameter)"
    )
    parser.add_argument("-f", "--filter", help="which filter to use (Otsu or binary)")
    parser.add_argument(
        "-l", "--lower", type=float, help="lower threshold for binary threshold"
    )
    parser.add_argument(
        "-u", "--upper", type=float, help="upper threshold for binary threshold"
    )
    parser.add_argument("file", help="file to compute circumference from")
    args = parser.parse_args()

    if args.debug:
        cli_settings.DEBUG = True
        print("Debug CLI option supplied.")

    cli_settings.RAW = args.raw

    if args.x:
        cli_settings.THETA_X = args.x

    if args.y:
        cli_settings.THETA_Y = args.y

    if args.z:
        cli_settings.THETA_Z = args.z

    if args.slice:
        cli_settings.SLICE = args.slice

    if args.conductance:
        cli_settings.CONDUCTANCE_PARAMETER = args.conductance

    if args.iterations:
        cli_settings.SMOOTHING_ITERATIONS = args.iterations

    if args.step:
        cli_settings.TIME_STEP = args.step

    if args.filter:
        if args.filter.lower() == "otsu":
            cli_settings.THRESHOLD_FILTER = constants.ThresholdFilter.Otsu
        elif args.filter.lower() == "binary":
            if args.lower is None:
                print(
                    "Must specify lower binary threshold with CLI option if using binary threshold"
                )
                exit(1)
            if args.upper is None:
                print(
                    "Must specify upper binary threshold with CLI option if using binary threshold"
                )
                exit(1)
            cli_settings.THRESHOLD_FILTER = constants.ThresholdFilter.Binary
        else:
            print("Invalid setting entered for CLI filter option.")
            exit(1)

    if args.lower:
        cli_settings.LOWER_BINARY_THRESHOLD = args.lower

    if args.upper:
        cli_settings.UPPER_BINARY_THRESHOLD = args.upper

    # Would use pathlib.Path .suffix here
    # However, '.nii.gz'.suffix would return '.gz'
    # But we expect '.nii.gz'
    extension = args.file[args.file.find(".") :]
    if extension not in constants.SUPPORTED_IMAGE_EXTENSIONS:
        print(
            f"Image has an unsupported extension. Supported image extensions are {str(constants.SUPPORTED_IMAGE_EXTENSIONS)}."
        )
        exit(1)

    cli_settings.FILE = args.file


def parse_gui_cli() -> None:
    """Parse GUI CLI args and set settings in ``gui_settings.py``.

    :return: None"""
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", help="print debug info", action="store_true")
    parser.add_argument(
        "-e",
        "--export-index",
        help="exported file names use the index displayed in the GUI instead of the \
                        original file name",
        action="store_true",
    )
    parser.add_argument(
        "-t",
        "--theme",
        help="configure theme, options are "
        + list_of_options_to_str(constants.THEMES).replace('"', ""),
    )
    parser.add_argument(
        "-c",
        "--color",
        help="contour color as name (e.g. red) or hex color code rrggbb",
    )
    args = parser.parse_args()

    if args.debug:
        gui_settings.DEBUG = True
        print("Debug CLI option supplied.")

    if args.export_index:
        gui_settings.EXPORTED_FILE_NAMES_USE_INDEX = True
        print("Exported files will use the index displayed in the GUI.")

    if args.theme:
        if args.theme not in constants.THEMES:
            print(
                f"Invalid theme specified. Options are {list_of_options_to_str(constants.THEMES)}"
            )
            exit(1)

        gui_settings.THEME_NAME = args.theme
        gui_settings.CONTOUR_COLOR = parse_main_color_from_theme_json()
        print(f"Theme {args.theme} specified.")

    if args.color:
        gui_settings.CONTOUR_COLOR = args.color
        print(
            f"Contour color is {'#' if not args.color.isalpha() else ''}{args.color}."
        )


def parse_cli_config() -> None:
    """Parse CLI JSON config and set user settings in gui_settings.py.

    load_json will load constants.JSON_CLI_CONFIG_PATH."""
    global JSON_SETTINGS
    JSON_SETTINGS = load_json(constants.JSON_CLI_CONFIG_PATH)
    cli_settings.DEBUG = parse_bool("DEBUG")
    if cli_settings.DEBUG:
        print("Printing debug messages.")

    cli_settings.RAW = parse_bool("RAW")
    cli_settings.THETA_X = parse_int("X")
    cli_settings.THETA_Y = parse_int("Y")
    cli_settings.THETA_Z = parse_int("Z")
    cli_settings.SLICE = parse_int("SLICE")
    cli_settings.CONDUCTANCE_PARAMETER = parse_float("CONDUCTANCE")
    cli_settings.SMOOTHING_ITERATIONS = parse_int("SMOOTHING")
    cli_settings.TIME_STEP = parse_float("TIME_STEP")
    if parse_str("THRESHOLD_FILTER").lower() == "otsu":
        cli_settings.THRESHOLD_FILTER = constants.ThresholdFilter.Otsu
    elif parse_str("THRESHOLD_FILTER").lower() == "binary":
        cli_settings.THRESHOLD_FILTER = constants.ThresholdFilter.Binary
    else:
        raise exceptions.InvalidJSONField("THRESHOLD_FILTER", "str")
    cli_settings.LOWER_BINARY_THRESHOLD = parse_float("LOWER_BINARY_THRESHOLD")
    cli_settings.UPPER_BINARY_THRESHOLD = parse_float("UPPER_BINARY_THRESHOLD")


def parse_gui_config() -> None:
    """Parse GUI JSON config and set user settings in gui_settings.py.

    load_json will load constants.JSON_GUI_CONFIG_PATH."""
    global JSON_SETTINGS
    JSON_SETTINGS = load_json(constants.JSON_GUI_CONFIG_PATH)

    gui_settings.DEBUG = parse_bool("DEBUG")
    if gui_settings.DEBUG:
        print("Printing debug messages.")
    gui_settings.FILE_BROWSER_START_DIR = parse_path("FILE_BROWSER_START_DIR")
    gui_settings.EXPORTED_FILE_NAMES_USE_INDEX = parse_bool(
        "EXPORTED_FILE_NAMES_USE_INDEX"
    )

    gui_settings.THEME_NAME = JSON_SETTINGS["THEME_NAME"]
    if gui_settings.THEME_NAME not in constants.THEMES:
        raise exceptions.InvalidJSONField(
            "THEME_NAME", list_of_options_to_str(constants.THEMES)
        )

    contour_color: str = JSON_SETTINGS["CONTOUR_COLOR"]
    if contour_color == "":
        gui_settings.CONTOUR_COLOR = parse_main_color_from_theme_json()
    # A name, e.g. red, green, blue. etc., which can be converted to a QColor
    elif contour_color.isalpha():
        gui_settings.CONTOUR_COLOR = contour_color
    elif len(contour_color) == 6 and all(
        char in string.hexdigits for char in contour_color
    ):
        gui_settings.CONTOUR_COLOR = contour_color
    else:
        raise exceptions.InvalidJSONField(
            "CONTOUR_COLOR",
            'Name (e.g., "red", "blue") or 6-hexit color code "RRGGBB" or "" (empty string) '
            "to set a default color based on theme",
        )

    gui_settings.STARTUP_WIDTH_RATIO = parse_float("STARTUP_WIDTH_RATIO")
    gui_settings.STARTUP_HEIGHT_RATIO = parse_float("STARTUP_HEIGHT_RATIO")
    gui_settings.DISPLAY_ADVANCED_MENU_MESSAGES_IN_TERMINAL = parse_bool(
        "DISPLAY_ADVANCED_MENU_MESSAGES_IN_TERMINAL"
    )


def parse_main_color_from_theme_json() -> str:
    """Parse the main color from the theme JSON file (user_settings.THEME_NAME) in the highlight field.

    Uses user_settings.THEME_NAME so must be called after parse_gui_config sets user_settings.THEME_NAME
    (i.e. can be called within parse_gui_config).

    :return: main color rrggbb (hexits)
    :rtype: str"""
    path_to_theme_json: Path = (
        constants.THEME_DIR
        / gui_settings.THEME_NAME
        / (gui_settings.THEME_NAME + ".json")
    )
    theme_json: dict = load_json(path_to_theme_json)
    color: str = theme_json["highlight"]
    if len(color) == 7:
        return color[1:]
    else:
        if "rgba(" not in color:
            raise Exception(
                f"{path_to_theme_json} has an invalid highlight field. Must be #rrggbb or rgba(r, g, b, a) (decimal)"
            )
        color = color.replace("rgba(", "").replace(")", "")
        channels: list[str] = color.split(",")
        r, g, b = int(channels[0]), int(channels[1]), int(channels[2])
        r, g, b = hex(r)[2:].zfill(2), hex(g)[2:].zfill(2), hex(b)[2:].zfill(2)
        return r + g + b


# Source: https://github.com/Alexhuszagh/BreezeStyleSheets/blob/main/configure.py#L82
def load_json(path: Path) -> dict:
    """Load config.json file, ignoring comments //.

    Source: https://github.com/Alexhuszagh/BreezeStyleSheets/blob/main/configure.py#L82

    :param path: path to JSON configuration file
    :type path: Path
    :return: JSON represented as dict
    :rtype: dict
    """
    with open(path) as f:
        lines = f.read().splitlines()
    lines = [i for i in lines if not i.strip().startswith("//")]
    return json.loads("\n".join(lines))


def parse_str(field: str) -> str:
    """For str field, return the str.

    :param field: JSON field
    :type field: ````str````
    :return: string
    :rtype: ````str````"""
    try:
        return JSON_SETTINGS[field]
    except:
        raise exceptions.InvalidJSONField(field, "str")


def parse_bool(field: str) -> bool:
    """For bool field "True" or "False", return the bool.

    :param field: JSON field
    :type field: str
    :raise: exceptions.InvalidJSONField if s is not "True" or "False"
    :return: True or False
    :rtype: bool"""
    if JSON_SETTINGS[field] != "True" and JSON_SETTINGS[field] != "False":
        raise exceptions.InvalidJSONField(
            field, 'bool ("True" or "False", with quotation marks)'
        )
    return True if JSON_SETTINGS[field] == "True" else False


def parse_path(field: str) -> Path:
    """For path field, return a Path.

    :param field: JSON field
    :type field: str
    :raise: exceptions.InvalidJSONField
    :return:
    :rtype: Path"""
    try:
        return Path(JSON_SETTINGS[field])
    except:
        raise exceptions.InvalidJSONField(field, "path")


def parse_int(field: str) -> int:
    """For int field, return int.

    :param field: JSON field
    :type field: str
    :raise: exceptions.InvalidJSONField
    :return:
    :rtype: int"""
    try:
        return int(JSON_SETTINGS[field])
    except:
        raise exceptions.InvalidJSONField(field, "int")


def parse_float(field: str) -> float:
    """For float field, return float.

    :param field: JSON field
    :type field: str
    :raise: exceptions.InvalidJSONField
    :return:
    :rtype: float"""
    try:
        return float(JSON_SETTINGS[field])
    except:
        raise exceptions.InvalidJSONField(field, "float")


def list_of_options_to_str(strs: list[str]) -> str:
    r"""Convert list[str] of options of the form ['a', 'b', 'c', 'd'] to str representation
    "a", "b", "c", or "d"

    :param strs:
    :type strs: list[str]
    :return: str representation "a", "b", "c", or "d"
    :rtype: str"""
    # Need at least 1 comma in the list
    assert len(strs) > 2
    s = str(strs)[1:-1].replace("'", '"')
    final_comma_position: int = s.rfind(",")
    return s[: final_comma_position + 1] + " or " + s[final_comma_position + 2 :]



"""Test that ``unittest`` tests are discovered when running ``pytest`` and ``tox`` (run in GitHub CI).

The failed test cases caused failures in CI when not skipped, as expected."""

import unittest


class TrivialFailures(unittest.TestCase):
    def test_pass(self):
        self.assertEqual(1, 1)

    @unittest.skip("This is expected to fail")
    def test_fail(self):
        self.fail()

    @unittest.skip("This is expected to fail")
    def test_error(self):
        raise RuntimeError("This is an error")



"""Test algorithm by computing R^2 from GUI calculation and ground truth data.

Simulates GUI button clicks.

Uses GUI. GUI imports and tests will not run in CI. See note in tests/README.md."""

from tests.constants import (
    TARGET_R_SQUARED,
    labeled_result,
    UBUNTU_GITHUB_ACTIONS_CI,
    compute_r_squared,
)
import os
import sys
import unittest
import pytest

if not UBUNTU_GITHUB_ACTIONS_CI:
    from PyQt6.QtWidgets import QApplication, QPushButton
    from PyQt6.QtTest import QTest
    from PyQt6.QtCore import Qt
    from NeuroRuler.utils.img_helpers import *
    from NeuroRuler.GUI.main import *


class MainWindowTest(unittest.TestCase):
    """Test the program by simulating GUI actions."""

    @unittest.skipIf(
        UBUNTU_GITHUB_ACTIONS_CI,
        reason="No GUI on Ubuntu GitHub Actions CI environment",
    )
    def test_alg(self):
        app = QApplication(sys.argv)
        self.form = (
            MainWindow()
        )  # Pass the local parent object to the child constructor

        labeled_data = []
        calculated_data = []
        data_folder = "data"

        length = 0

        for file_name in os.listdir(data_folder):
            if file_name.endswith("w.nrrd"):
                nrrd_prefix = os.path.splitext(file_name)[0][
                    :10
                ]  # Remove the file extension to get the prefix
                for tsv_file_name in os.listdir(data_folder):
                    if nrrd_prefix in tsv_file_name and tsv_file_name.endswith(".tsv"):
                        length = length + 1

                        # Get labeled data
                        labeled_data.append(
                            labeled_result(data_folder + "/" + tsv_file_name)
                        )

                        # Simulate GUI actions
                        self.form.browse_files(False, data_folder + "/" + file_name)
                        self.form.toggle_setting_to_true()
                        button = QPushButton("apply")
                        button.clicked.connect(self.form.settings_export_view_toggle)
                        button.click()

                        # Get circumference
                        circumference_label: str = self.form.circumference_label
                        split_parts = circumference_label.text().split(":")
                        part_after_colon = split_parts[-1].strip()[:3]
                        calculated_data.append(float(part_after_colon))
                        break
                else:
                    continue

        self.assertTrue(length > 10)
        self.assertTrue(
            compute_r_squared(labeled_data, calculated_data) > TARGET_R_SQUARED
        )



from NeuroRuler.utils.img_helpers import *
import NeuroRuler.utils.global_vars as global_vars

IMAGE_NAMES: list[str] = [
    "BCP_Dataset_2month_T1w.nrrd",
    "IBIS_Case1_V06_t1w_RAI.nrrd",
    "IBIS_Case2_V12_t1w_RAI.nrrd",
    "IBIS_Case3_V24_t1w_RAI.nrrd",
    "IBIS_Dataset_12month_T1w.nrrd",  # Group 1: 0-4
    "IBIS_Dataset_NotAligned_6month_T1w.nrrd",  # Group 2: 5
    "MicroBiome_1month_T1w.nii.gz",  # Group 3: 6
]

IMAGE_PATHS: list[Path] = [constants.DATA_DIR / path_str for path_str in IMAGE_NAMES]

IMAGE_DICT: dict[Path, sitk.Image] = dict()

for path in IMAGE_PATHS:
    global_vars.READER.SetFileName(str(path))
    IMAGE_DICT[path] = global_vars.READER.Execute()

GROUP_1: list[sitk.Image] = [IMAGE_DICT[k] for k in IMAGE_PATHS[:5]]
GROUP_2: list[sitk.Image] = [IMAGE_DICT[IMAGE_PATHS[5]]]
GROUP_3: list[sitk.Image] = [IMAGE_DICT[IMAGE_PATHS[6]]]


def test_setup():
    clear_globals()


def test_hashing():
    """Test that the properties tuples for the 5 images in group 1 are all the same
    but different from the properties tuple for group 2 and group 3."""
    first_group_hash = hash(get_properties_from_sitk_image(GROUP_1[0]))
    second_group_hash = hash(get_properties_from_sitk_image(GROUP_2[0]))
    third_group_hash = hash(get_properties_from_sitk_image(GROUP_3[0]))
    assert all(
        hash(get_properties_from_sitk_image(img)) == first_group_hash for img in GROUP_1
    )
    assert second_group_hash != third_group_hash
    assert first_group_hash != second_group_hash
    assert first_group_hash != third_group_hash


def test_curr_path():
    clear_globals()
    initialize_globals(IMAGE_PATHS)
    for i in range(4):
        next_img()
    assert get_curr_path() == IMAGE_PATHS[4]
    for i in range(2):
        next_img()
    # Since the group is only the first 5
    # Wraps around
    assert get_curr_path() == IMAGE_PATHS[1]
    next_img()
    assert get_curr_path() == IMAGE_PATHS[2]


def test_del_curr_img_delete_last():
    clear_globals()
    initialize_globals(IMAGE_PATHS)
    global_vars.CURR_IMAGE_INDEX = len(global_vars.IMAGE_DICT) - 1
    del_curr_img()
    assert global_vars.CURR_IMAGE_INDEX == len(global_vars.IMAGE_DICT) - 1



"""CLI settings set through command-line arguments and ``cli_config.json``.

Command-line arguments override the values in the JSON."""

from typing import Any
import NeuroRuler.utils.global_vars as global_vars
from NeuroRuler.utils.constants import ThresholdFilter

DEBUG: bool = False
"""Whether or not to print debugging information throughout execution."""

RAW: bool = False
"""Whether to print the \"raw\" circumference.

If false, it prints a \"pretty\", rounded output with units included."""

FILE: str = ""
"""The file."""

THETA_X: int = global_vars.THETA_X
"""In degrees"""
THETA_Y: int = global_vars.THETA_Y
"""In degrees"""
THETA_Z: int = global_vars.THETA_Z
"""In degrees"""
SLICE: int = -1
"""0-indexed. Overwritten later."""

CONDUCTANCE_PARAMETER: float = global_vars.CONDUCTANCE_PARAMETER
"""Smoothing option. See global_vars.CONDUCTANCE_PARAMETER."""
SMOOTHING_ITERATIONS: int = global_vars.SMOOTHING_ITERATIONS
"""Smoothing option. See global_vars.SMOOTHING_ITERATIONS."""
TIME_STEP: float = global_vars.TIME_STEP
"""Smoothing option. See global_vars.TIME_STEP."""

THRESHOLD_FILTER: ThresholdFilter = ThresholdFilter.Otsu
"""Which threshold filter to use. Default is Otsu."""

LOWER_BINARY_THRESHOLD: float = global_vars.LOWER_BINARY_THRESHOLD
"""Threshold option for binary threshold."""
UPPER_BINARY_THRESHOLD: float = global_vars.UPPER_BINARY_THRESHOLD
"""Threshold option for binary threshold."""


def get_settings() -> dict[str, Any]:
    r"""Returns ``dict`` containing values of all variables in this file, used for debugging

    :return: ``dict`` of all CLI settings
    :rtype: ``dict[str, Any]``"""
    return {
        "DEBUG": DEBUG,
        "RAW": RAW,
        "FILE": FILE,
        "THETA_X": THETA_X,
        "THETA_Y": THETA_Y,
        "THETA_Z": THETA_Z,
        "SLICE": SLICE,
        "CONDUCTANCE_PARAMETER": CONDUCTANCE_PARAMETER,
        "SMOOTHING_ITERATIONS": SMOOTHING_ITERATIONS,
        "TIME_STEP": TIME_STEP,
        "THRESHOLD_FILTER": THRESHOLD_FILTER,
        "LOWER_BINARY_THRESHOLD": LOWER_BINARY_THRESHOLD,
        "UPPER_BINARY_THRESHOLD": UPPER_BINARY_THRESHOLD,
    }



"""GUI helper functions."""

import platform
import string
from typing import Union

import SimpleITK as sitk
import numpy as np

from PyQt6.QtGui import QImage, QColor, QPixmap, QIcon, QFont
from PyQt6 import QtWidgets
from PyQt6.QtWidgets import (
    QApplication,
    QDialog,
    QLabel,
    QMainWindow,
    QFileDialog,
    QMenu,
    QVBoxLayout,
    QWidget,
    QMessageBox,
)
from PyQt6.QtCore import QSize
from PyQt6.QtCore import Qt

import qimage2ndarray

import NeuroRuler.utils.exceptions as exceptions
import NeuroRuler.utils.gui_settings as user_settings
from NeuroRuler.utils.constants import deprecated

MACOS: bool = "macOS" in platform.platform()
WINDOW_TITLE_PADDING: int = 12
"""Used in InformationDialog to add width to the dialog to prevent the window title from being truncated."""


# tl;dr QColor can have alpha (e.g., if we wanted contour color to be transparent)
# but we don't have a need for it so don't support it.
# Call hasAlphaChannel() on many of the QImage's we're working with results in False.
# qimage2ndarray supports scalar/gray + alpha and RGB + alpha, but perhaps the numpy arrays
# we get from sitk.Image don't have alpha. We don't need to go to the effort of adding alpha.
def string_to_QColor(name_or_hex: str) -> QColor:
    """Convert a name (e.g. red) or 6-hexit rrggbb string to a ``QColor``.

    :param name_or_hex: name of color (e.g. blue) or rrggbb (hexits)
    :type name_or_hex: str
    :return: QColor
    :rtype: QColor
    :raise: exceptions.InvalidColor if ``name_or_hex`` not in specified formats"""
    if name_or_hex.isalpha():
        return QColor(name_or_hex)
    if not all(char in string.hexdigits for char in name_or_hex):
        raise exceptions.InvalidColor(name_or_hex)

    channels: bytes = bytes.fromhex(name_or_hex)
    if len(channels) == 3:
        return QColor(channels[0], channels[1], channels[2])
    else:
        raise exceptions.InvalidColor(name_or_hex)


def mask_QImage(q_img: QImage, binary_mask: np.ndarray, color: QColor) -> None:
    """Given 2D ``q_img`` and 2D ``binary_mask`` of the same shape, apply ``binary_mask`` on ``q_img``
    to change ``q_img`` pixels corresponding to ``binary_mask``=1 to ``color``. Mutates ``q_img``.

    QImage and numpy use [reversed w,h order](https://stackoverflow.com/a/68220805/18479243).

    This function checks that
    ``q_img.size().width() == binary_mask.shape[0]`` and ``q_img.size().height() == binary_mask.shape[1]``.

    :param q_img:
    :type q_img: QImage
    :param binary_mask: 0|1 elements
    :type binary_mask: np.ndarray
    :param color:
    :type color: QColor
    :raise: exceptions.ArraysDifferentShape if the arrays are of different shape
    :return: None
    :rtype: None"""
    if (
        q_img.size().width() != binary_mask.shape[0]
        or q_img.size().height() != binary_mask.shape[1]
    ):
        raise exceptions.ArraysDifferentShape
    for i in range(binary_mask.shape[0]):
        for j in range(binary_mask.shape[1]):
            if binary_mask[i][j]:
                q_img.setPixelColor(i, j, color)


def sitk_slice_to_qimage(sitk_slice: sitk.Image) -> QImage:
    """Convert a 2D sitk.Image slice to a QImage.

    This function calls sitk.GetArrayFromImage, which returns the transpose.
    It also calls qimage2ndarray.array2qimage with normalize=True, normalizing
    the pixels to 0..255.

    :param sitk_slice: 2D slice
    :type sitk_slice: sitk.Image
    :return: 0..255 normalized QImage
    :rtype: QImage"""
    slice_np: np.ndarray = sitk.GetArrayFromImage(sitk_slice)
    return qimage2ndarray.array2qimage(slice_np, normalize=True)


class ErrorMessageBox(QMessageBox):
    def __init__(self, message: str):
        """:param message: Error message
        :type message: str"""
        super().__init__()
        # Window title is ignored on macOS.
        # QDialog's window title is not ignored on macOS, but I'm pretty sure QDialog doesn't
        # support icons.
        self.setWindowTitle("Error")
        self.setIconPixmap(
            QPixmap(f":/{user_settings.THEME_NAME}/message_critical.svg")
        )
        self.setText(message)


# adjustSize adjusts size based on window content, not window title (+ window buttons)
# So for some menu options, the window title would be truncated if some width isn't added
# However, QDialog does show window title on macOS (unlike QMessageBox)
class InformationDialog(QDialog):
    def __init__(self, title: str, message: str):
        """:param title: Title of window
        :type title: str
        :param message: Informational message
        :type message: str"""
        super().__init__()
        self.setWindowTitle(title)
        layout: QVBoxLayout = QVBoxLayout()
        message_label: QLabel = QLabel(message)
        layout.addWidget(message_label)
        self.setLayout(layout)
        self.adjustSize()
        # Add width to prevent truncation of the window title
        self.setFixedSize(
            self.size().width() + WINDOW_TITLE_PADDING * len(title),
            self.size().height(),
        )


# Deprecated because QMessageBox's window title doesn't show up on macOS
# However, QMessageBox can display an icon, whereas QDialog can't (I think)
# The icon provides some additional width that ill cause the window title to not be truncated, unlike QDialog
@deprecated
class InformationMessageBox(QMessageBox):
    def __init__(self, title: str, message: str):
        """:param title: Title of window (on macOS, QMessageBox window title doesn't show up so is instead prepended to the message)
        :type title: str
        :param message: Informational message
        :type message: str"""
        super().__init__()
        self.setWindowTitle(title)
        self.setIconPixmap(
            QPixmap(f":/{user_settings.THEME_NAME}/message_information.svg")
        )
        if MACOS:
            title += "\n\n"
        self.setText(f"{title if MACOS else ''} {message}")



"""Defines ``MainWindow`` and ``main()``, the entrypoint of the GUI.

Loads ``NeuroRuler/GUI/mainwindow.ui``, made in QtDesigner.

Loads ``.qss`` stylesheets and ``resources.py`` (icons) files, generated
by BreezeStyleSheets. Our fork of the repo: https://github.com/NIRALUser/BreezeStyleSheets.

If adding a new GUI element (in the GUI or in the menubar, whatever), you'll have to modify
modify __init__ and settings_view_toggle.

Edge cases: If this element should be disabled after enable_elements or enabled after disable_elements,
then you will need to modify those."""


import importlib
import sys
import os
import json
import re
import webbrowser
from pathlib import Path
from typing import Union

import SimpleITK as sitk
import numpy as np
from typing import Any

from PyQt6 import QtGui, QtCore
from PyQt6.QtGui import QPixmap, QAction, QImage, QIcon, QResizeEvent
from PyQt6.QtWidgets import (
    QApplication,
    QDialog,
    QLabel,
    QMainWindow,
    QFileDialog,
    QMenu,
    QVBoxLayout,
    QWidget,
    QMessageBox,
)
from PyQt6.uic.load_ui import loadUi
from PyQt6.QtCore import Qt, QSize

import pprint
import pkg_resources
from NeuroRuler.utils.constants import View, ThresholdFilter
import NeuroRuler.utils.constants as constants

# Note, do not use imports like
# from NeuroRuler.utils.global_vars import IMAGE_DICT
# This would make the global variables not work
import NeuroRuler.utils.global_vars as global_vars
import NeuroRuler.utils.imgproc as imgproc
import NeuroRuler.utils.gui_settings as settings
from NeuroRuler.GUI.helpers import (
    string_to_QColor,
    mask_QImage,
    sitk_slice_to_qimage,
    ErrorMessageBox,
    InformationDialog,
)

from NeuroRuler.utils.img_helpers import (
    initialize_globals,
    update_images,
    get_curr_image,
    get_curr_image_size,
    get_curr_rotated_slice,
    get_curr_smooth_slice,
    get_curr_metadata,
    get_curr_binary_thresholded_slice,
    get_curr_otsu_slice,
    get_curr_physical_units,
    get_curr_path,
    get_all_paths,
    get_curr_properties_tuple,
    get_middle_dimension,
)

import NeuroRuler.utils.img_helpers as img_helpers


PATH_TO_UI_FILE: Path = Path("NeuroRuler") / "GUI" / "mainwindow.ui"
if not PATH_TO_UI_FILE.exists():
    PATH_TO_UI_FILE = Path(
        pkg_resources.resource_filename("NeuroRuler.GUI", "mainwindow.ui")
    )
PATH_TO_NR_LOGO: Path = Path("NeuroRuler") / "GUI" / "static" / "nr_logo.png"
if not PATH_TO_NR_LOGO.exists():
    PATH_TO_NR_LOGO = Path(
        pkg_resources.resource_filename("NeuroRuler.GUI", "static/nr_logo.png")
    )

SETTINGS_VIEW_ENABLED: bool = True
"""Whether the user is able to adjust settings (settings screen) or not
(circumference and contoured image screen)."""

DEFAULT_CIRCUMFERENCE_LABEL_TEXT: str = "Calculated Circumference: N/A"
DEFAULT_IMAGE_PATH_LABEL_TEXT: str = "Image path"
GITHUB_LINK: str = "https://github.com/NIRALUser/NeuroRuler"
DOCUMENTATION_LINK: str = "https://NeuroRuler.readthedocs.io/en/latest/"
DEFAULT_IMAGE_TEXT: str = "Select images using File > Open!"
DEFAULT_IMAGE_NUM_LABEL_TEXT: str = "Image 0 of 0"
DEFAULT_IMAGE_STATUS_TEXT: str = "Image path is displayed here."

UNSCALED_QPIXMAP: QPixmap
"""Unscaled QPixmap from which the scaled version is rendered in the GUI.

When any slice (rotated, smoothed, previewed) is rendered from an unscaled QImage, this variable is set to the
QPixmap generated from that unscaled QImage.

This variable will not change on resizeEvent. resizeEvent will scale this. Otherwise, if scaling
self.image's pixmap (which is already scaled), there would be loss of detail."""
OUTPUT_SLICE_EXTENSION: str = "png"


class MainWindow(QMainWindow):
    """Main window of the application.

    Settings mode and circumference mode."""

    def __init__(self):
        """Load main file and connect GUI events to methods/functions.

        Sets window title and icon."""
        super(MainWindow, self).__init__()
        loadUi(str(PATH_TO_UI_FILE), self)

        self.setWindowTitle("NeuroRuler")

        self.action_open.triggered.connect(lambda: self.browse_files(False))
        self.action_add_images.triggered.connect(lambda: self.browse_files(True))
        self.action_remove_image.triggered.connect(self.remove_curr_img)
        self.action_exit.triggered.connect(exit)
        self.action_github.triggered.connect(lambda: webbrowser.open(GITHUB_LINK))
        self.action_documentation.triggered.connect(
            lambda: webbrowser.open(DOCUMENTATION_LINK)
        )
        self.action_show_credits.triggered.connect(
            lambda: information_dialog(
                "Credits",
                'Credit to Jesse Wei, Madison Lester, Peifeng "Hank" He, Eric Schneider, and Martin Styner.\n\nUniversity of North Carolina at Chapel Hill, 2023. See the GitHub page for more info.',
            )
        )
        self.action_show_dimensions.triggered.connect(display_dimensions)
        self.action_show_properties.triggered.connect(display_properties)
        self.action_show_direction.triggered.connect(display_direction)
        self.action_show_spacing.triggered.connect(display_spacing)
        self.action_export_json.triggered.connect(self.export_json)
        self.action_export_png.triggered.connect(
            lambda: self.export_curr_slice_as_img("png")
        )
        self.action_export_jpg.triggered.connect(
            lambda: self.export_curr_slice_as_img("jpg")
        )
        self.action_export_bmp.triggered.connect(
            lambda: self.export_curr_slice_as_img("bmp")
        )
        self.action_export_ppm.triggered.connect(
            lambda: self.export_curr_slice_as_img("ppm")
        )
        self.action_export_xbm.triggered.connect(
            lambda: self.export_curr_slice_as_img("xbm")
        )
        self.action_export_xpm.triggered.connect(
            lambda: self.export_curr_slice_as_img("xpm")
        )
        self.action_import.triggered.connect(self.import_json)
        self.next_button.clicked.connect(self.next_img)
        self.previous_button.clicked.connect(self.previous_img)
        self.apply_button.clicked.connect(self.settings_export_view_toggle)
        self.x_slider.valueChanged.connect(self.rotate_x)
        self.y_slider.valueChanged.connect(self.rotate_y)
        self.z_slider.valueChanged.connect(self.rotate_z)
        self.slice_slider.valueChanged.connect(self.slice_update)
        self.reset_button.clicked.connect(self.reset_settings)
        self.smoothing_preview_button.clicked.connect(self.render_smooth_slice)
        self.otsu_radio_button.clicked.connect(self.disable_binary_threshold_inputs)
        self.binary_radio_button.clicked.connect(self.enable_binary_threshold_inputs)
        self.threshold_preview_button.clicked.connect(self.render_threshold)
        self.x_view_radio_button.clicked.connect(self.update_view)
        self.y_view_radio_button.clicked.connect(self.update_view)
        self.z_view_radio_button.clicked.connect(self.update_view)

        self.export_button.clicked.connect(self.export_json)

    def enable_elements(self) -> None:
        """Called after File > Open.

        Enables GUI elements. Explicitly disables some (e.g., Export CSV menu item and
        binary threshold inputs, since Otsu is default).

        :return: None
        """
        # findChildren searches recursively by default
        for widget in self.findChildren(QWidget):
            widget.setEnabled(True)

        # Menu stuff
        for widget in self.findChildren(QAction):
            widget.setEnabled(True)

        self.action_export_json.setEnabled(not SETTINGS_VIEW_ENABLED)
        self.export_button.setEnabled(not SETTINGS_VIEW_ENABLED)
        self.disable_binary_threshold_inputs()

    def enable_binary_threshold_inputs(self) -> None:
        """Called when Binary filter button is clicked.

        Restore binary input box.

        :return: None
        """
        self.upper_threshold_input.setEnabled(True)
        self.lower_threshold_input.setEnabled(True)

    def settings_export_view_toggle(self) -> None:
        """Called when clicking Apply (in settings mode) or Adjust (in circumference mode).

        Toggle SETTINGS_VIEW_ENABLED, change apply button text, render stuff depending on the current mode.

        Enables/disables GUI elements depending on the value of SETTINGS_VIEW_ENABLED.

        :return: None
        """
        # Unsure sure why this is necessary here but nowhere else...
        global SETTINGS_VIEW_ENABLED
        SETTINGS_VIEW_ENABLED = not SETTINGS_VIEW_ENABLED
        settings_view_enabled = SETTINGS_VIEW_ENABLED
        if settings_view_enabled:
            self.apply_button.setText("Apply")
            self.circumference_label.setText(DEFAULT_CIRCUMFERENCE_LABEL_TEXT)
            # Render uncontoured slice after pressing adjust
            self.render_curr_slice()
        else:
            self.apply_button.setText("Adjust")

            # When clicking Apply, the circumference should be calculated only for the axial slice.
            # Edge case: If the user does a huge rotation, then switching to axial view may not display an
            # axial slice. For example, after clicking coronal and setting X rotation to 90,
            # clicking axial will not show an axial slice.
            # We could just not change the view and re-orient when clicking Apply if this is a valid use case (probably isn't).
            # TODO: Check with Styner
            self.set_view_z()
            self.orient_curr_image()

            self.update_smoothing_settings(True)
            self.update_binary_filter_settings(True)
            # Ignore the type annotation warning here.
            # render_curr_slice() must return np.ndarray since not settings_view_enabled here
            binary_contour_slice: np.ndarray = self.render_curr_slice()
            self.render_circumference(binary_contour_slice)

        # Open button is always enabled.
        # If pressing it in circumference mode, then browse_files() will toggle to settings view.
        self.action_open.setEnabled(True)
        self.action_import.setEnabled(settings_view_enabled)
        self.action_add_images.setEnabled(settings_view_enabled)
        self.action_remove_image.setEnabled(settings_view_enabled)
        self.x_slider.setEnabled(settings_view_enabled)
        self.y_slider.setEnabled(settings_view_enabled)
        self.z_slider.setEnabled(settings_view_enabled)
        self.slice_slider.setEnabled(settings_view_enabled)
        self.x_rotation_label.setEnabled(settings_view_enabled)
        self.y_rotation_label.setEnabled(settings_view_enabled)
        self.z_rotation_label.setEnabled(settings_view_enabled)
        self.slice_num_label.setEnabled(settings_view_enabled)
        self.reset_button.setEnabled(settings_view_enabled)
        self.smoothing_preview_button.setEnabled(settings_view_enabled)
        self.otsu_radio_button.setEnabled(settings_view_enabled)
        self.binary_radio_button.setEnabled(settings_view_enabled)
        self.lower_threshold.setEnabled(settings_view_enabled)
        self.lower_threshold_input.setEnabled(settings_view_enabled)
        self.upper_threshold.setEnabled(settings_view_enabled)
        self.upper_threshold_input.setEnabled(settings_view_enabled)
        self.threshold_preview_button.setEnabled(settings_view_enabled)
        self.action_export_json.setEnabled(not settings_view_enabled)
        self.circumference_label.setEnabled(not settings_view_enabled)
        self.export_button.setEnabled(not settings_view_enabled)
        self.smoothing_preview_button.setEnabled(settings_view_enabled)
        self.conductance_parameter_label.setEnabled(settings_view_enabled)
        self.conductance_parameter_input.setEnabled(settings_view_enabled)
        self.smoothing_iterations_label.setEnabled(settings_view_enabled)
        self.smoothing_iterations_input.setEnabled(settings_view_enabled)
        self.time_step_label.setEnabled(settings_view_enabled)
        self.time_step_input.setEnabled(settings_view_enabled)
        self.x_view_radio_button.setEnabled(settings_view_enabled)
        self.y_view_radio_button.setEnabled(settings_view_enabled)
        self.z_view_radio_button.setEnabled(settings_view_enabled)
        self.lower_threshold_input.setEnabled(
            settings_view_enabled and self.binary_radio_button.isChecked()
        )
        self.upper_threshold_input.setEnabled(
            settings_view_enabled and self.binary_radio_button.isChecked()
        )

    def disable_binary_threshold_inputs(self) -> None:
        """Called when Otsu filter button is clicked.

        Disable binary threshold input boxes.

        :return: None
        """
        self.upper_threshold_input.setEnabled(False)
        self.lower_threshold_input.setEnabled(False)

    def disable_elements(self) -> None:
        """Called when the list is now empty, i.e. just removed from list of length 1.

        Explicitly enables elements that should never be disabled and sets default text.

        :return: None
        """
        central_widget = self.findChildren(QWidget, "centralwidget")[0]
        menubar = self.menuBar()

        for gui_element in central_widget.findChildren(QWidget):
            gui_element.setEnabled(False)

        # findChildren searches recursively by default
        for menu in menubar.findChildren(QMenu):
            for action in menu.actions():
                action.setEnabled(False)

        self.action_open.setEnabled(True)
        self.circumference_label.setText(DEFAULT_CIRCUMFERENCE_LABEL_TEXT)
        self.image.setEnabled(True)
        self.image.clear()
        self.image.setText(DEFAULT_IMAGE_TEXT)
        self.image.setStatusTip(DEFAULT_IMAGE_STATUS_TEXT)
        self.image_path_label.setText(DEFAULT_IMAGE_PATH_LABEL_TEXT)
        self.image_num_label.setText(DEFAULT_IMAGE_NUM_LABEL_TEXT)
        self.apply_button.setText("Apply")
        self.z_view_radio_button.setChecked(True)

    def browse_files(self, extend: bool, path=None) -> None:
        """Called after File > Open or File > Add Images.

        If ``extend``, then ``IMAGE_DICT`` will be updated with new images.

        Else, ``IMAGE_DICT`` will be cleared and
        (re)initialized (e.g. when choosing files for the first time or re-opening).

        Opens file menu.

        Renders various elements depending on the value of ``extend``.

        If called in circumference mode, then will toggle to settings mode.

        :param extend: Whether to clear IMAGE_DICT and (re)initialize or add images to it. Determines which GUI elements are rendered.
        :param path: Used for unit testing, when only one path is imported. Normally, the path(s) are selected by user in a QFileDialog.
        :type extend: bool
        :return: None"""
        # If called in circumference mode, then toggle to settings mode.
        if not SETTINGS_VIEW_ENABLED:
            self.settings_export_view_toggle()

        if path is None:
            file_filter: str = "MRI Images ("
            for extension in constants.SUPPORTED_IMAGE_EXTENSIONS:
                file_filter += "*" + extension + " "
            # Get rid of trailing whitespace
            file_filter = file_filter[:-1]
            file_filter += ")"

            files = QFileDialog.getOpenFileNames(
                self,
                "Open files",
                str(settings.FILE_BROWSER_START_DIR),
                file_filter,
            )
            # list[str]
            path_list = files[0]
            if len(path_list) == 0:
                return
        else:
            path_list = [path]

        # Convert to list[Path]. Slight inefficiency but worth.
        path_list = list(map(Path, path_list))

        differing_images: list[Path]

        if not extend:
            differing_images = initialize_globals(path_list)
            # Set view to z because initialize_globals calls update_images, which orients loaded images
            # for the axial view
            self.set_view_z()
            self.render_all_sliders()
            self.enable_elements()
            self.render_curr_slice()
            if differing_images:
                newline: str = "\n"
                error_message_box(
                    f"The image(s) you uploaded have differing properties.\n"
                    f"The first one and all images with properties matching the first one have been loaded.\n"
                    f"The name(s) of the ones with differing properties are\n\n"
                    f"{newline.join([path.name for path in differing_images])}"
                )
        else:
            # Doesn't need to re-render sliders to set max value of slice slider.
            # update_images won't change max value of slice slicer.
            # Does not need to render current slice. Images are added to the end of the dict.
            # And adding duplicate key doesn't change key order.
            differing_images = update_images(path_list)
            if differing_images:
                newline: str = "\n"
                error_message_box(
                    f"You have uploaded image(s) with properties that differ from those of the currently loaded ones.\n"
                    f"These image(s) have not been loaded:\n\n"
                    f"{newline.join([path.name for path in differing_images])}"
                )
        # When extending, image num must be updated
        self.render_image_num_and_path()

    def update_view(self) -> None:
        """Called when clicking on any of the three view radio buttons.

        Sets global_vars.VIEW to the correct value. Then orients the current image and renders.

        :return: None
        """
        # The three buttons are in a button group in the  file
        # And all have autoExclusive=True
        if self.x_view_radio_button.isChecked():
            global_vars.VIEW = constants.View.X
        elif self.y_view_radio_button.isChecked():
            global_vars.VIEW = constants.View.Y
        else:
            global_vars.VIEW = constants.View.Z

        self.orient_curr_image()
        self.render_curr_slice()

    def set_view_z(self) -> None:
        """Set global_vars.VIEW to View.Z and set the z radio button to checked.

        :return: None
        """
        global_vars.VIEW = constants.View.Z
        # TODO: Uncheck x and y are technically unnecessary since these 3 buttons in the view_button_group have
        # autoExclusive=True
        self.x_view_radio_button.setChecked(False)
        self.y_view_radio_button.setChecked(False)
        self.z_view_radio_button.setChecked(True)

    def update_smoothing_settings(self, set_global_vars_to_GUI_text: bool) -> None:
        """Update smoothing text in the GUI and set SMOOTHING_FILTER parameters.

        :param set_global_vars_to_GUI_text: If True, will first try to modify global_vars variables to the text in the GUI before updating GUI text and filter parameters. If False, will not do so.
        :type set_global_vars_to_GUI_text: bool
        :return: None
        """
        if set_global_vars_to_GUI_text:
            try:
                global_vars.CONDUCTANCE_PARAMETER = float(
                    self.conductance_parameter_input.displayText()
                )
            except ValueError:
                if settings.DEBUG:
                    print("Conductance must be a float!")
        self.conductance_parameter_input.setText(str(global_vars.CONDUCTANCE_PARAMETER))
        self.conductance_parameter_input.setPlaceholderText(
            str(global_vars.CONDUCTANCE_PARAMETER)
        )
        global_vars.SMOOTHING_FILTER.SetConductanceParameter(
            global_vars.CONDUCTANCE_PARAMETER
        )

        if set_global_vars_to_GUI_text:
            try:
                global_vars.SMOOTHING_ITERATIONS = int(
                    self.smoothing_iterations_input.displayText()
                )
            except ValueError:
                if settings.DEBUG:
                    print("Iterations must be an integer!")
        self.smoothing_iterations_input.setText(str(global_vars.SMOOTHING_ITERATIONS))
        self.smoothing_iterations_input.setPlaceholderText(
            str(global_vars.SMOOTHING_ITERATIONS)
        )
        global_vars.SMOOTHING_FILTER.SetNumberOfIterations(
            global_vars.SMOOTHING_ITERATIONS
        )

        if set_global_vars_to_GUI_text:
            try:
                global_vars.TIME_STEP = float(self.time_step_input.displayText())
            except ValueError:
                if settings.DEBUG:
                    print("Time step must be a float!")
        self.time_step_input.setText(str(global_vars.TIME_STEP))
        self.time_step_input.setPlaceholderText(str(global_vars.TIME_STEP))
        global_vars.SMOOTHING_FILTER.SetTimeStep(global_vars.TIME_STEP)

    def update_binary_filter_settings(self, set_global_vars_to_GUI_text: bool) -> None:
        """Updates binary threshold filter text in the GUI and set BINARY_THRESHOLD_FILTER parameters.

        :param set_global_vars_to_GUI_text: If True, will first try to modify global_vars variables to the text in the GUI before updating GUI text and filter parameters. If False, will not do so.
        :type set_global_vars_to_GUI_text: bool
        :return: None
        """
        if set_global_vars_to_GUI_text:
            try:
                global_vars.LOWER_BINARY_THRESHOLD = float(
                    self.lower_threshold_input.displayText()
                )
            except ValueError:
                pass
        self.lower_threshold_input.setText(str(global_vars.LOWER_BINARY_THRESHOLD))
        self.lower_threshold_input.setPlaceholderText(
            str(global_vars.LOWER_BINARY_THRESHOLD)
        )
        global_vars.BINARY_THRESHOLD_FILTER.SetLowerThreshold(
            global_vars.LOWER_BINARY_THRESHOLD
        )

        if set_global_vars_to_GUI_text:
            try:
                global_vars.UPPER_BINARY_THRESHOLD = float(
                    self.upper_threshold_input.displayText()
                )
            except ValueError:
                pass
        self.upper_threshold_input.setText(str(global_vars.UPPER_BINARY_THRESHOLD))
        self.upper_threshold_input.setPlaceholderText(
            str(global_vars.UPPER_BINARY_THRESHOLD)
        )
        global_vars.BINARY_THRESHOLD_FILTER.SetUpperThreshold(
            global_vars.UPPER_BINARY_THRESHOLD
        )

    def render_scaled_qpixmap_from_qimage(self, q_img: QImage) -> None:
        """Convert q_img to QPixmap and set self.image's pixmap to that pixmap scaled to self.image's size.

        Sets UNSCALED_PIXMAP to the unscaled pixmap generated from the q_img.

        :param q_img:
        :type q_img: QImage
        :return: None"""
        global UNSCALED_QPIXMAP
        UNSCALED_QPIXMAP = QPixmap(q_img)
        self.image.setPixmap(
            UNSCALED_QPIXMAP.scaled(
                self.image.size(),
                aspectRatioMode=Qt.AspectRatioMode.KeepAspectRatio,
                transformMode=Qt.TransformationMode.SmoothTransformation,
            )
        )

    def resizeEvent(self, event: QResizeEvent) -> None:
        """This method is called every time the window is resized. Overrides PyQt6's resizeEvent.

        Sets pixmap to UNSCALED_QPIXMAP scaled to self.image's size.

        :param event:
        :type event: QResizeEvent
        :return: None"""
        if global_vars.IMAGE_DICT:
            self.image.setPixmap(
                UNSCALED_QPIXMAP.scaled(
                    self.image.size(),
                    aspectRatioMode=Qt.AspectRatioMode.KeepAspectRatio,
                    transformMode=Qt.TransformationMode.SmoothTransformation,
                )
            )
        QMainWindow.resizeEvent(self, event)

    def render_curr_slice(self) -> Union[np.ndarray, None]:
        """Resamples the currently selected image using its rotation and slice settings,
        then renders the resulting slice (scaled to the size of self.image) in the GUI.

        DOES NOT set text for ``image_num_label`` and file path labels.

        If ``not SETTINGS_VIEW_ENABLED``, also calls ``imgproc.contour()`` and outlines
        the contour of the QImage (mutating it).

        Additionally, also returns a view of the binary contoured slice if ``not SETTINGS_VIEW_ENABLED``.
        This saves work when computing circumference.

        :return: np.ndarray if ``not SETTINGS_VIEW_ENABLED`` else None
        :rtype: np.ndarray or None"""

        if not SETTINGS_VIEW_ENABLED:
            self.set_view_z()

        rotated_slice: sitk.Image = get_curr_rotated_slice()
        q_img: QImage = sitk_slice_to_qimage(rotated_slice)
        rv_dummy_var: np.ndarray = np.zeros(0)

        if not SETTINGS_VIEW_ENABLED:
            if self.otsu_radio_button.isChecked():
                binary_contour_slice: np.ndarray = imgproc.contour(
                    rotated_slice, ThresholdFilter.Otsu
                )
            else:
                binary_contour_slice: np.ndarray = imgproc.contour(
                    rotated_slice, ThresholdFilter.Binary
                )
            rv_dummy_var = binary_contour_slice
            mask_QImage(
                q_img,
                np.transpose(binary_contour_slice),
                string_to_QColor(settings.CONTOUR_COLOR),
            )

        elif global_vars.VIEW != constants.View.Z:
            z_indicator: np.ndarray = np.zeros(
                (rotated_slice.GetSize()[1], rotated_slice.GetSize()[0])
            )
            z_indicator[get_curr_image_size()[2] - global_vars.SLICE - 1, :] = 1
            mask_QImage(
                q_img,
                np.transpose(z_indicator),
                string_to_QColor(settings.CONTOUR_COLOR),
            )

        self.render_scaled_qpixmap_from_qimage(q_img)

        if not SETTINGS_VIEW_ENABLED:
            return rv_dummy_var

    def render_smooth_slice(self) -> None:
        """Renders smooth slice in GUI. Allows user to preview result of smoothing settings.

        :return: None"""
        self.update_smoothing_settings(True)
        # Preview should apply filter only on axial slice
        self.set_view_z()
        smooth_slice: sitk.Image = get_curr_smooth_slice()
        q_img: QImage = sitk_slice_to_qimage(smooth_slice)
        self.render_scaled_qpixmap_from_qimage(q_img)

    def render_threshold(self) -> None:
        """Render filtered image slice on UI.

        :return: None"""
        # Preview should apply filter only on axial slice
        self.set_view_z()
        if self.otsu_radio_button.isChecked():
            filter_img: sitk.Image = get_curr_otsu_slice()
        else:
            self.update_binary_filter_settings(True)
            filter_img: sitk.Image = get_curr_binary_thresholded_slice()
        q_img: QImage = sitk_slice_to_qimage(filter_img)
        self.render_scaled_qpixmap_from_qimage(q_img)

    def render_circumference(self, binary_contour_slice: np.ndarray) -> float:
        """Called after pressing Apply or when
        (not SETTINGS_VIEW_ENABLED and (pressing Next or Previous or Remove Image))

        Computes circumference from binary_contour_slice and renders circumference label.

        binary_contour_slice is always the return value of render_curr_slice since render_curr_slice must have
        already been called. If calling this function, render_curr_slice must have been called first.

        :param binary_contour_slice: Result of previously calling render_curr_slice when ``not SETTINGS_VIEW_ENABLED``
        :type binary_contour_slice: np.ndarray
        :return: circumference
        :rtype: float"""
        if SETTINGS_VIEW_ENABLED:
            raise Exception("Rendering circumference label when SETTINGS_VIEW_ENABLED")
        units: Union[str, None] = get_curr_physical_units()

        # Euler3D rotation has no effect on spacing (see unit test). This is the correct spacing
        # This is also the same as get_curr_rotated_slice().GetSpacing(), just without index [2]
        spacing: tuple = get_curr_image().GetSpacing()

        if settings.DEBUG:
            print(f"Computing circumference, and this is the spacing: {spacing}")

        # TODO
        # binary_contour_slice is the transpose of the rotated_slice
        # Thus, should pass spacing values in the reverse order?
        circumference: float = imgproc.length_of_contour_with_spacing(
            binary_contour_slice, spacing[0], spacing[1]
        )
        # circumference: float = imgproc.length_of_contour(binary_contour_slice)
        self.circumference_label.setText(
            f"Calculated Circumference: {round(circumference, constants.NUM_DIGITS_TO_ROUND_TO)} {units if units is not None else constants.MESSAGE_TO_SHOW_IF_UNITS_NOT_FOUND}"
        )
        return circumference

    def toggle_setting_to_false(self) -> None:
        """Used in testing.

        Flipping the SETTINGS_VIEW_ENABLED

        :return: None"""
        global SETTINGS_VIEW_ENABLED
        if SETTINGS_VIEW_ENABLED:
            SETTINGS_VIEW_ENABLED = not SETTINGS_VIEW_ENABLED

    def toggle_setting_to_true(self) -> None:
        """Used in testing.

        Flipping the SETTINGS_VIEW_ENABLED

        :return: None"""
        global SETTINGS_VIEW_ENABLED
        if not SETTINGS_VIEW_ENABLED:
            SETTINGS_VIEW_ENABLED = not SETTINGS_VIEW_ENABLED

    def render_image_num_and_path(self) -> None:
        """Set image_num_label, image_path_label, and status tip of the image.

        Called when pressing Next or Previous (next_img, prev_img), and after File > Open (browse_files).

        Also called when removing an image.

        :return: None"""
        self.image_num_label.setText(
            f"Image {global_vars.CURR_IMAGE_INDEX + 1} of {len(global_vars.IMAGE_DICT)}"
        )
        self.image_path_label.setText(str(get_curr_path().name))
        self.image_path_label.setStatusTip(str(get_curr_path()))
        self.image.setStatusTip(str(get_curr_path()))

    def render_all_sliders(self) -> None:
        """Sets all slider values to the global rotation and slice values.
        Also updates maximum value of slice slider.

        Called on reset. Will need to be called when updating batch index, if we implement this.

        Not called when the user updates a slider.

        Also updates rotation and slice num labels.

        :return: None"""
        self.x_slider.setValue(global_vars.THETA_X)
        self.y_slider.setValue(global_vars.THETA_Y)
        self.z_slider.setValue(global_vars.THETA_Z)
        self.slice_slider.setMaximum(get_curr_image().GetSize()[View.Z.value] - 1)
        self.slice_slider.setValue(global_vars.SLICE)
        self.x_rotation_label.setText(f"X rotation: {global_vars.THETA_X}°")
        self.y_rotation_label.setText(f"Y rotation: {global_vars.THETA_Y}°")
        self.z_rotation_label.setText(f"Z rotation: {global_vars.THETA_Z}°")
        self.slice_num_label.setText(f"Slice: {global_vars.SLICE}")

    def rotate_x(self) -> None:
        """Called when the user updates the x slider.

        Render image and set ``x_rotation_label``.

        :return: None"""
        x_slider_val: int = self.x_slider.value()
        global_vars.THETA_X = x_slider_val
        self.render_curr_slice()
        self.x_rotation_label.setText(f"X rotation: {x_slider_val}°")

    def rotate_y(self) -> None:
        """Called when the user updates the y slider.

        Render image and set ``y_rotation_label``.

        :return: None"""
        y_slider_val: int = self.y_slider.value()
        global_vars.THETA_Y = y_slider_val
        self.render_curr_slice()
        self.y_rotation_label.setText(f"Y rotation: {y_slider_val}°")

    def rotate_z(self) -> None:
        """Called when the user updates the z slider.

        Render image and set ``z_rotation_label``.

        :return: None"""
        z_slider_val: int = self.z_slider.value()
        global_vars.THETA_Z = z_slider_val
        self.render_curr_slice()
        self.z_rotation_label.setText(f"Z rotation: {z_slider_val}°")

    def slice_update(self) -> None:
        """Called when the user updates the slice slider.

        Render image and set ``slice_num_label``.

        :return: None"""
        slice_slider_val: int = self.slice_slider.value()
        global_vars.SLICE = slice_slider_val
        self.render_curr_slice()
        self.slice_num_label.setText(f"Slice: {slice_slider_val}")

    def reset_settings(self) -> None:
        """Called when Reset is clicked.

        Resets rotation values to 0 and slice num to the default ``int((z-1)/2)``
        for the current image, then renders current image and sliders.

        :return: None"""
        global_vars.THETA_X = 0
        global_vars.THETA_Y = 0
        global_vars.THETA_Z = 0
        global_vars.SLICE = get_middle_dimension(get_curr_image(), View.Z)
        self.render_curr_slice()
        self.render_all_sliders()

    def next_img(self) -> None:
        """Called when Next button is clicked.

        Advance index and render.

        :return: None"""
        img_helpers.next_img()
        # TODO: This feels inefficient...
        self.orient_curr_image()
        binary_contour_or_none: Union[np.ndarray, None] = self.render_curr_slice()
        self.render_image_num_and_path()

        if not SETTINGS_VIEW_ENABLED:
            # Ignore the type annotation warning. binary_contour_or_none must be binary_contour since not SETTINGS_VIEW_ENABLED
            self.render_circumference(binary_contour_or_none)

    def previous_img(self) -> None:
        """Called when Previous button is clicked.

        Decrement index and render.

        :return: None"""
        img_helpers.previous_img()
        # TODO: This feels inefficient...
        self.orient_curr_image()
        binary_contour_or_none: Union[np.ndarray, None] = self.render_curr_slice()
        self.render_image_num_and_path()

        if not SETTINGS_VIEW_ENABLED:
            # Ignore the type annotation warning. binary_contour_or_none must be binary_contour since not SETTINGS_VIEW_ENABLED
            self.render_circumference(binary_contour_or_none)

    # TODO: Due to the images now being a dict, we can
    # easily let the user remove a range of images if they want
    def remove_curr_img(self) -> None:
        """Called after File > Remove File.

        Removes current image from ``IMAGE_DICT``. Since ``IMAGE_DICT`` is a reference to an image dict
        in ``IMAGE_GROUPS``, it's removed from ``IMAGE_GROUPS`` as well.

        :return: None"""
        img_helpers.del_curr_img()

        if len(global_vars.IMAGE_DICT) == 0:
            self.disable_elements()
            return

        binary_contour_or_none: Union[np.ndarray, None] = self.render_curr_slice()
        self.render_image_num_and_path()

        if not SETTINGS_VIEW_ENABLED:
            # Ignore the type annotation warning. binary_contour_or_none must be binary_contour since not SETTINGS_VIEW_ENABLED
            self.render_circumference(binary_contour_or_none)

    def test_stuff(self) -> None:
        """Connected to Debug > Test stuff. Dummy button and function for easily testing stuff.

        Assume that anything you put here will be overwritten freely.

        :return: None"""
        self.image.setPixmap(QPixmap(f":/{settings.THEME_NAME}/help.svg"))
        self.image.setStatusTip(
            "This is intentional, if it's a question mark then that's good :), means we can display icons"
        )

    # TODO: File name should also include circumference when not SETTINGS_VIEW_ENABLED?
    def export_curr_slice_as_img(self, extension: str) -> None:
        """Called when an Export as image menu item is clicked.

        Exports ``self.image`` to ``constants.OUTPUT_DIR/image_stem/``. Thus, calling this when ``SETTINGS_VIEW_ENABLED`` will
        save a non-contoured image. Calling this when ``not SETTINGS_VIEW_ENABLED`` will save a contoured
        image.

        Filename has format <file_name>[_contoured].<extension>

        _contoured will be in the name if ``not SETTINGS_VIEW_ENABLED``.

        Supported formats in this function are the ones supported by QPixmap,
        namely BMP, JPG, JPEG, PNG, PPM, XBM, XPM.

        :param extension: BMP, JPG, JPEG, PNG, PPM, XBM, XPM
        :type extension: str
        :param path:
        :type path: Path
        :return: ``None``"""
        file_stem: str = get_curr_path().stem
        output_path: Path = constants.OUTPUT_DIR / file_stem

        if not output_path.exists():
            output_path.mkdir()

        path: str = str(
            output_path
            / f"{file_stem}{'_contoured' if not SETTINGS_VIEW_ENABLED else ''}.{extension}"
        )
        self.image.pixmap().save(path, extension)

    def import_json(self) -> None:
        """Called when "import" button is clicked

        Imported parameters include: input_image_path, output_folder, x_rotation, y_rotation, z_rotation, slice,
        smoothing_conductance, smoothing_iterations, smoothing_time_step, filter_option, upper_binary_threshold, lower_binary_threshold,
        and circumference

        input_image_path is the only mandatory field.

        :return: `None`"""
        file_filter: str = "NeuroRuler MRI image JSON " + str(("*.json")).replace(
            "'", ""
        ).replace(",", "")

        files, _ = QFileDialog.getOpenFileNames(
            self, "Open file", str(settings.FILE_BROWSER_START_DIR), file_filter
        )

        # list[str]
        path_list = files
        if len(path_list) == 0:
            return
        elif len(path_list) > 1:
            information_dialog(
                "Multiple imports",
                "Multiple JSON files were selected. Only the first will be imported.",
            )

        with open(path_list[0], "r") as file:
            # Parse the JSON data into a dictionary
            data = json.load(file)

        image_path: str = data["input_image_path"]
        self.browse_files(False, image_path)

        if "x_rotation" in data:
            global_vars.THETA_X = data["x_rotation"]
            self.x_slider.setValue(global_vars.THETA_X)

        if "y_rotation" in data:
            global_vars.THETA_Y = data["y_rotation"]
            self.y_slider.setValue(global_vars.THETA_Y)

        if "z_rotation" in data:
            global_vars.THETA_Z = data["z_rotation"]
            self.z_slider.setValue(global_vars.THETA_Z)

        self.update_view()

        if "slice" in data:
            global_vars.SLICE = data["slice"]
            self.slice_slider.setValue(global_vars.SLICE)
            self.slice_update()

        if "smoothing_conductance" in data:
            global_vars.CONDUCTANCE_PARAMETER = data["smoothing_conductance"]

        if "smoothing_iterations" in data:
            global_vars.SMOOTHING_ITERATIONS = data["smoothing_iterations"]

        if "smoothing_time_step" in data:
            global_vars.TIME_STEP = data["smoothing_time_step"]

        self.update_smoothing_settings(False)

        if "filter_option" in data:
            if (
                not "upper_binary_threshold" in data
                or not "lower_binary_threshold" in data
            ):
                print(
                    "If the imported JSON has filter_option (whether Otsu or binary), it must have upper_binary_threshold and lower_binary_threshold"
                )
                exit(1)
            global_vars.UPPER_BINARY_THRESHOLD = data["upper_binary_threshold"]
            global_vars.LOWER_BINARY_THRESHOLD = data["lower_binary_threshold"]
            self.update_binary_filter_settings(False)

            if data["filter_option"] == "Otsu":
                self.otsu_radio_button.click()
                self.disable_binary_threshold_inputs()
            else:
                self.enable_binary_threshold_inputs()
                self.binary_radio_button.click()

    def export_json(self) -> None:
        """Called when "export" button is clicked and when Menu > Export > JSON is clicked.

        Exported parameters include: image_name, output_folder, x_rotation, y_rotation, z_rotation, slice,
        smoothing_conductance, smoothing_iterations, smoothing_time_step, filter_option, upper_binary_threshold, lower_binary_threshold,
        and circumference

        :return: `None`"""

        global_vars.CURR_IMAGE_INDEX = 0
        self.render_image_num_and_path()
        for image_num in range(len(global_vars.IMAGE_DICT)):
            curr_path: Path = get_curr_path()
            stem: str = curr_path.stem
            binary_contour_slice: np.ndarray = self.render_curr_slice()
            circumference: float = self.render_circumference(binary_contour_slice)
            output_dir: Path = constants.OUTPUT_DIR / stem
            if not output_dir.exists():
                output_dir.mkdir()

            self.export_curr_slice_as_img(OUTPUT_SLICE_EXTENSION)

            data: dict[str, Any] = {
                "input_image_path": str(curr_path),
                "output_contoured_slice_path": str(
                    Path.cwd()
                    / output_dir
                    / (stem + f"_contoured.{OUTPUT_SLICE_EXTENSION}")
                ),
                "x_rotation": global_vars.THETA_X,
                "y_rotation": global_vars.THETA_Y,
                "z_rotation": global_vars.THETA_Z,
                "slice": global_vars.SLICE,
                "smoothing_conductance": global_vars.CONDUCTANCE_PARAMETER,
                "smoothing_iterations": global_vars.SMOOTHING_ITERATIONS,
                "smoothing_time_step": global_vars.TIME_STEP,
                "filter_option": "Otsu"
                if self.otsu_radio_button.isChecked()
                else "Binary",
                "upper_binary_threshold": global_vars.UPPER_BINARY_THRESHOLD,
                "lower_binary_threshold": global_vars.LOWER_BINARY_THRESHOLD,
                "circumference": circumference,
            }

            with open(output_dir / (stem + "_settings.json"), "w") as outfile:
                json.dump(data, outfile, indent=4)

            self.next_img()

    def orient_curr_image(self) -> None:
        """Orient the current image for the current view (global_vars.VIEW) by applying ORIENT_FILTER on it.

        This mutates the image.

        :return: None"""
        img_helpers.orient_curr_image(global_vars.VIEW)


def error_message_box(message: str) -> None:
    """Creates a message box with an error message and red warning icon.

    :param message: the error message to be displayed
    :type message: str
    :return: None"""
    ErrorMessageBox(message).exec()


def information_dialog(title: str, message: str) -> None:
    """Create an informational dialog QDialog window with title and message.

    :param title:
    :type title: str
    :param message:
    :type message: str
    :return: None"""
    InformationDialog(title, message).exec()


# TODO: Broken
def display_metadata() -> None:
    """Display metadata in window or terminal. Internally, uses sitk.GetMetaData, which doesn't return
    all metadata (e.g., doesn't return spacing values whereas sitk.GetSpacing does).

    Typically, this returns less metadata for NRRD than for NIfTI.

    :return: None"""
    if not len(global_vars.IMAGE_DICT):
        print("Can't print metadata when there's no image!")
        return
    message: str = pprint.pformat(get_curr_metadata())
    if settings.DISPLAY_ADVANCED_MENU_MESSAGES_IN_TERMINAL:
        print(message)
    else:
        information_dialog("Metadata", message)


def display_dimensions() -> None:
    """Display current image's dimensions in window or terminal.

    :return: None"""
    if not len(global_vars.IMAGE_DICT):
        print("Can't print dimensions when there's no image!")
        return
    message: str = pprint.pformat(get_curr_image().GetSize())
    if settings.DISPLAY_ADVANCED_MENU_MESSAGES_IN_TERMINAL:
        print(message)
    else:
        information_dialog("Dimensions", message)


# TODO: If updating img_helpers.get_properties(), this needs to be slightly adjusted!
def display_properties() -> None:
    """Display properties in window or terminal.

    Internally, the properties tuple is a tuple of values only and doesn't contain
    field names. This function creates a dictionary with field names for printing. But the dictionary
    doesn't exist in the program.

    :return: None"""
    if not len(global_vars.IMAGE_DICT):
        print("No loaded image!")
        return
    curr_properties: tuple = get_curr_properties_tuple()
    fields: tuple = ("center of rotation", "dimensions", "spacing")
    if len(fields) != len(curr_properties):
        print(
            "Update NeuroRuler/GUI/main.print_properties() !\nNumber of fields and number of properties don't match."
        )
        exit(1)
    # Pretty sure the dict(zip(...)) goes through fields in alphabetical order
    message: str = pprint.pformat(dict(zip(fields, curr_properties)))
    if settings.DISPLAY_ADVANCED_MENU_MESSAGES_IN_TERMINAL:
        print(message)
    else:
        information_dialog("Properties", message)


def display_direction() -> None:
    """Display current image's direction in window or terminal.

    :return: None"""
    if not len(global_vars.IMAGE_DICT):
        print("Can't print direction when there's no image!")
        return
    message: str = pprint.pformat(get_curr_image().GetDirection())
    if settings.DISPLAY_ADVANCED_MENU_MESSAGES_IN_TERMINAL:
        print(message)
    else:
        information_dialog("Direction", message)


def display_spacing() -> None:
    """Display current image's spacing in window or terminal.

    :return: None"""
    if not len(global_vars.IMAGE_DICT):
        print("Can't print spacing when there's no image!")
        return
    message: str = pprint.pformat(get_curr_image().GetSpacing())
    if settings.DISPLAY_ADVANCED_MENU_MESSAGES_IN_TERMINAL:
        print(message)
    else:
        information_dialog("Spacing", message)


def main() -> None:
    """Main entrypoint of GUI."""
    # This import can't go at the top of the file
    # because gui.py.parse_gui_cli() has to set THEME_NAME before the import occurs
    # This imports globally
    # For example, NeuroRuler/GUI/helpers.py can access resource files without having to import there
    importlib.import_module(f"NeuroRuler.GUI.themes.{settings.THEME_NAME}.resources")

    app = QApplication(sys.argv)

    # On macOS, sets the application logo in the dock (but no window icon on macOS)
    # TODO
    # On Windows, sets the window icon at the top left of the window (but no dock icon on Windows)
    app.setWindowIcon(QIcon(str(PATH_TO_NR_LOGO)))

    # TODO: Put arrow buttons on the left and right endpoints of the sliders
    # These arrow buttons already show up if commenting in app.setStyle("Fusion")
    # And commenting out with open stylesheet and app.setStyleSheet
    # We should figure out how to get arrow buttons on sliders for (+, -) 1 precise adjustments.
    # Currently, the sliders allow this (left click on the left or right end), but the arrow buttons
    # are not in the GUI.
    # app.setStyle("Fusion")

    MAIN_WINDOW: MainWindow = MainWindow()

    with open(constants.THEME_DIR / settings.THEME_NAME / "stylesheet.qss", "r") as f:
        MAIN_WINDOW.setStyleSheet(f.read())

    # Non-zero min width and height is needed to prevent
    # this bug https://github.com/NIRALUser/NeuroRuler/issues/42
    # However, this also seems to affect startup GUI size or at least GUI element spacing
    MAIN_WINDOW.setMinimumSize(QSize(1, 1))
    MAIN_WINDOW.resize(
        int(settings.STARTUP_WIDTH_RATIO * constants.PRIMARY_MONITOR_DIMENSIONS[0]),
        int(settings.STARTUP_HEIGHT_RATIO * constants.PRIMARY_MONITOR_DIMENSIONS[1]),
    )

    MAIN_WINDOW.show()

    try:
        # sys.exit will cause a bug when running from terminal
        # After importing the GUI runner function from __init__, clicking the close window button
        # (not the menu button) will not close the window
        # because the Python process wouldn't end
        os._exit(app.exec())
    except:
        if settings.DEBUG:
            print("Exiting")


if __name__ == "__main__":
    import NeuroRuler.utils.parser as parser

    parser.parse_gui_config()
    parser.parse_gui_cli()
    main()



"""Global variables that change throughout program execution."""

import SimpleITK as sitk
from pathlib import Path
from typing import Any
from NeuroRuler.utils.constants import View

IMAGE_DICT: dict[Path, sitk.Image] = dict()
"""The group of images that have been loaded.

Since Python 3.7+, dicts maintain insertion order. Therefore, we can use CURR_IMAGE_INDEX for retrieval and deletion.

Use list(IMAGE_DICT.keys())[i] to return the i'th key in the dict, which can also index into the dict.
Not sure about this operation's speed, but it's used only
in the GUI for insertion and deletion operations, should be fine.

All images in the dictionary have matching properties, as defined by mri_image.get_properties."""

CURR_IMAGE_INDEX: int = 0
"""Image of the current image in the loaded batch of images, which is a dict[Path, sitk.Image].

You should probably use the helper functions in img_helpers instead of this (unless you're writing helper
functions)."""

READER: sitk.ImageFileReader = sitk.ImageFileReader()
"""Global ``sitk.ImageFileReader``."""

ORIENT_FILTER: sitk.DICOMOrientImageFilter = sitk.DICOMOrientImageFilter()
"""Global sitk.DICOMOrientImageFilter for orienting images.

See https://simpleitk.org/doxygen/latest/html/classitk_1_1simple_1_1DICOMOrientImageFilter.html#details
and the orientation strings in constants.py. Use ITK-SNAP for the orientations that we copy."""
VIEW: View = View.Z
"""Current view in the GUI."""
X_CENTER: int = 0
"""Used for changing views."""
Y_CENTER: int = 0
"""Used for changing views."""

EULER_3D_TRANSFORM: sitk.Euler3DTransform = sitk.Euler3DTransform()
"""Global sitk.Euler3DTransform for 3D rotations."""
THETA_X: int = 0
"""In degrees"""
THETA_Y: int = 0
"""In degrees"""
THETA_Z: int = 0
"""In degrees"""
SLICE: int = 0
"""0-indexed"""

SMOOTHING_FILTER: sitk.GradientAnisotropicDiffusionImageFilter = (
    sitk.GradientAnisotropicDiffusionImageFilter()
)
"""Global sitk.GradientAnisotropicDiffusionImageFilter for image smoothing.

See https://slicer.readthedocs.io/en/latest/user_guide/modules/gradientanisotropicdiffusion.html
for more information."""
CONDUCTANCE_PARAMETER: float = 3.0
"""Smoothing option.

Conductance controls the sensitivity of the conductance term.
As a general rule, the lower the value, the more strongly the filter preserves edges.
A high value will cause diffusion (smoothing) across edges.
Note that the number of iterations controls how much smoothing is done within regions bounded by edges."""
SMOOTHING_ITERATIONS: int = 5
"""Smoothing option.

The more iterations, the more smoothing. Each iteration takes the same amount of time.
If it takes 10 seconds for one iteration, then it will take 100 seconds for 10 iterations.
Note that the conductance controls how much each iteration smooths across edges."""
TIME_STEP: float = 0.0625
"""Smoothing option.

The time step depends on the dimensionality of the image.
In Slicer, the images are 3D and the default (.0625) time step will provide a stable solution."""

OTSU_THRESHOLD_FILTER: sitk.OtsuThresholdImageFilter = sitk.OtsuThresholdImageFilter()
"""Global Otsu threshold filter."""

BINARY_THRESHOLD_FILTER: sitk.BinaryThresholdImageFilter = (
    sitk.BinaryThresholdImageFilter()
)
"""Global binary threshold filter."""
LOWER_BINARY_THRESHOLD: float = 0.0
"""Threshold option for binary threshold."""
UPPER_BINARY_THRESHOLD: float = 200.0
"""Threshold option for binary threshold."""



"""Image helper functions that aren't quite part of the main algorithm, unlike ``imgproc.py``.

Mostly holds helper functions for working with ``IMAGE_DICT`` in ``global_vars.py``."""

from typing import Union
import SimpleITK as sitk
from pathlib import Path
import NeuroRuler.utils.global_vars as global_vars
from NeuroRuler.utils.constants import degrees_to_radians, View
import NeuroRuler.utils.constants as constants


def update_images(path_list: list[Path]) -> list[Path]:
    """Initialize IMAGE_DICT. See the docstring for IMAGE_DICT in global_vars.py for more info.

    All images are oriented for the axial view when loaded. When calling this, make sure global_vars.VIEW = Z.

    If the images at path(s) in path_list don't match the properties of previously saved images,
    then this method returns the paths of the images that don't match, and
    IMAGE_DICT is updated only with the non-differing images.

    :param path_list:
    :type path_list: list[Path]
    :raise: Exception if path_list is empty (disallow this in GUI)
    :return: List of paths of images with properties that differ from those of the images currently in IMAGE_DICT
    :rtype: list[Path]"""
    if not path_list:
        raise Exception("update_images assumes path_list isn't empty.")

    # On load, orient the image for Z view by default
    # If we don't do this, then the misaligned image's GetSize()[2] won't actually be the inferior-superior axis
    # Then the max slice value would not be correct because it would use some other axis
    global_vars.ORIENT_FILTER.SetDesiredCoordinateOrientation(
        constants.Z_ORIENTATION_STR
    )

    comparison_properties_tuple: tuple
    if global_vars.IMAGE_DICT:
        comparison_properties_tuple = get_curr_properties_tuple()
    else:
        first_path: Path = path_list[0]
        comparison_properties_tuple = get_properties_from_path(first_path)
        global_vars.READER.SetFileName(str(first_path))
        first_img: sitk.Image = global_vars.READER.Execute()
        first_img_axial: sitk.Image = global_vars.ORIENT_FILTER.Execute(first_img)
        global_vars.IMAGE_DICT[first_path] = first_img_axial
        # Don't need to look at first image again
        path_list = path_list[1:]

    differing_image_paths: list[Path] = []

    for path in path_list:
        global_vars.READER.SetFileName(str(path))
        new_img: sitk.Image = global_vars.READER.Execute()
        new_img = global_vars.ORIENT_FILTER.Execute(new_img)
        new_img_properties: tuple = get_properties_from_sitk_image(new_img)
        if comparison_properties_tuple != new_img_properties:
            differing_image_paths.append(path)
        else:
            new_img_axial: sitk.Image = global_vars.ORIENT_FILTER.Execute(new_img)
            global_vars.IMAGE_DICT[path] = new_img_axial
    return differing_image_paths


def initialize_globals(path_list: list[Path]) -> list[Path]:
    """After pressing File > Open, the global variables need to be cleared and (re)initialized.

    If loading images with different properties, then this method returns
    False, and IMAGE_DICT isn't updated with the differing images.

    Mutated global variables: IMAGE_DICT, CURR_IMAGE_INDEX,
    READER, THETA_X, THETA_Y, THETA_Z, SLICE, EULER_3D_TRANSFORM.

    Specifically, clears IMAGE_DICT and then populates it.

    :param path_list:
    :type path_list: list[Path]
    :return: List of paths of images with properties that differ from those of the images currently in IMAGE_DICT
    :rtype: list[Path]"""
    global_vars.CURR_IMAGE_INDEX = 0
    global_vars.IMAGE_DICT.clear()
    differing_image_paths: list[Path] = update_images(path_list)
    global_vars.THETA_X = 0
    global_vars.THETA_Y = 0
    global_vars.THETA_Z = 0
    # curr_img has the same properties as the whole group in IMAGE_DICT.
    # Use it to set slice and center of rotation for the group
    curr_img: sitk.Image = get_curr_image()
    # TODO: Set maximum for the slice slider in the GUI!
    global_vars.SLICE = get_middle_dimension(curr_img, View.Z)
    global_vars.EULER_3D_TRANSFORM.SetCenter(get_center_of_rotation(curr_img))
    global_vars.EULER_3D_TRANSFORM.SetRotation(0, 0, 0)
    global_vars.SMOOTHING_FILTER.SetConductanceParameter(3.0)
    global_vars.SMOOTHING_FILTER.SetNumberOfIterations(5)
    global_vars.SMOOTHING_FILTER.SetTimeStep(0.0625)
    global_vars.VIEW = constants.View.Z
    global_vars.X_CENTER = get_middle_dimension(curr_img, View.X)
    global_vars.Y_CENTER = get_middle_dimension(curr_img, View.Y)
    global_vars.BINARY_THRESHOLD_FILTER.SetLowerThreshold(100)
    global_vars.BINARY_THRESHOLD_FILTER.SetUpperThreshold(200)
    return differing_image_paths


def clear_globals() -> None:
    """Clear global variables for unit testing in test_img_helpers.

    Don't need to reset Euler3DTransform since that's not used in the tests there.

    :return: None
    :rtype: None"""
    global_vars.IMAGE_DICT.clear()
    global_vars.CURR_IMAGE_INDEX = 0
    global_vars.THETA_X = 0
    global_vars.THETA_Y = 0
    global_vars.THETA_Z = 0
    global_vars.SLICE = 0


def image_dict_is_empty() -> bool:
    """
    :return: True if IMAGE_DICT is empty, else False
    :rtype: bool"""
    return bool(global_vars.IMAGE_DICT)


def get_curr_path() -> Path:
    """Return the current Path in IMAGE_DICT. That is, the key at index CURR_IMAGE_INDEX.

    :return: Path of current image
    :rtype: Path"""
    return list(global_vars.IMAGE_DICT.keys())[global_vars.CURR_IMAGE_INDEX]


def get_all_paths() -> list[Path]:
    """Return all Paths in IMAGE_DICT.

    :return: All paths in IMAGE_DICT
    :rtype: list[Path]"""
    return list(global_vars.IMAGE_DICT.keys())


def get_curr_image() -> sitk.Image:
    """Return the sitk.Image at the current index in global_vars.IMAGE_DICT.

    :return: current image
    :rtype: sitk.Image"""
    return global_vars.IMAGE_DICT[get_curr_path()]


# TODO: Add more properties?
# TODO: Implement tolerance for spacing
def get_properties_from_sitk_image(img: sitk.Image) -> tuple:
    """Tuple of properties of a sitk.Image.

    TODO: Add more properties

    If modifying this, NeuroRuler/GUI/main.py print_properties also has to be modified slightly.

    :param img:
    :type img: sitk.Image
    :return: (center of rotation used in EULER_3D_TRANSFORM, dimensions, spacing)
    :rtype: tuple"""
    return (
        get_center_of_rotation(img),
        img.GetSize(),
        img.GetSpacing(),
    )


def get_properties_from_path(path: Path) -> tuple:
    """Tuple of properties of the sitk.Image we get from path. Uses global_vars.READER.

    :param path: Path from which we can get a sitk.Image
    :type path: Path
    :return: (dimensions, center of rotation used in EULER_3D_TRANSFORM, spacing)
    :rtype: tuple"""
    global_vars.READER.SetFileName(str(path))
    img: sitk.Image = global_vars.READER.Execute()
    return get_properties_from_sitk_image(img)


def get_middle_dimension(img: sitk.Image, axis: View) -> int:
    """int((img.GetSize()[axis.value] - 1) / 2)

    :param img:
    :param axis: int (0-2)
    :type img: sitk.Image
    :return: int((img.GetSize()[axis.value] - 1) / 2)
    :rtype: int"""
    return int((img.GetSize()[axis.value] - 1) / 2)


def get_curr_image_size() -> tuple:
    """Return dimensions of current image.

    :return: dimensions
    :rtype: tuple"""
    return get_curr_image().GetSize()


def get_center_of_rotation(img: sitk.Image) -> tuple:
    """img.TransformContinuousIndexToPhysicalPoint([
    (dimension - 1) / 2.0 for dimension in img.GetSize()])

    :param: img
    :type img: sitk.Image
    :return: img.TransformContinuousIndexToPhysicalPoint([
    (dimension - 1) / 2.0 for dimension in img.GetSize()]"""
    return img.TransformContinuousIndexToPhysicalPoint(
        [(dimension - 1) / 2.0 for dimension in img.GetSize()]
    )


def set_curr_image(image: sitk.Image) -> None:
    """Set the sitk.Image at the current index in global_vars.IMAGE_DICT.

    :param image:
    :type image: sitk.Image
    :return: None
    :rtype: None"""
    global_vars.IMAGE_DICT[get_curr_path()] = image


def orient_curr_image(view: View) -> None:
    """Given a view enum, set the current image to the oriented version for that view.

    :param view:
    :type view: View.X, View.Y, or View.Z
    :return: None
    :rtype: None"""
    if not isinstance(view, View):
        raise Exception(
            "Expected View.X, View.Y, or View.Z but did not get one of those."
        )
    global_vars.ORIENT_FILTER.SetDesiredCoordinateOrientation(
        constants.ORIENTATION_STRINGS[view.value]
    )
    oriented: sitk.Image = global_vars.ORIENT_FILTER.Execute(get_curr_image())
    set_curr_image(oriented)


def get_curr_rotated_slice() -> sitk.Image:
    """Return 2D rotated slice of the current image determined by global rotation and slice settings.

    Sets global_vars.EULER_3D_TRANSFORM's rotation values but not its center since all loaded images should
    have the same center.

    :return: 2D rotated slice
    :rtype: sitk.Image"""
    global_vars.EULER_3D_TRANSFORM.SetRotation(
        degrees_to_radians(global_vars.THETA_X),
        degrees_to_radians(global_vars.THETA_Y),
        degrees_to_radians(global_vars.THETA_Z),
    )
    rotated_image: sitk.Image = sitk.Resample(
        get_curr_image(), global_vars.EULER_3D_TRANSFORM
    )
    rotated_slice: sitk.Image
    if global_vars.VIEW == constants.View.X:
        rotated_slice = rotated_image[global_vars.X_CENTER, :, :]
    elif global_vars.VIEW == constants.View.Y:
        rotated_slice = rotated_image[:, global_vars.Y_CENTER, :]
    else:
        rotated_slice = rotated_image[:, :, global_vars.SLICE]

    return rotated_slice


def get_curr_smooth_slice() -> sitk.Image:
    """Return smoothed 2D rotated slice of the current image determined by global smoothing settings.

    :return: smooth 2D rotated slice
    :rtype: sitk.Image"""
    rotated_slice: sitk.Image = get_curr_rotated_slice()
    # The cast is necessary, otherwise get sitk::ERROR: Pixel type: 16-bit signed integer is not supported in 2D
    smooth_slice: sitk.Image = global_vars.SMOOTHING_FILTER.Execute(
        sitk.Cast(rotated_slice, sitk.sitkFloat64)
    )
    return smooth_slice


def get_curr_binary_thresholded_slice() -> sitk.Image:
    """Return binary thresholded slice of the current image based on binary filter determined by
    global threshold settings.

    :return: binary thresholded 2D rotated slice using global binary threshold settings
    :rtype: sitk.Image"""
    rotated_slice: sitk.Image = get_curr_rotated_slice()
    filter_slice: sitk.Image = global_vars.BINARY_THRESHOLD_FILTER.Execute(
        sitk.Cast(rotated_slice, sitk.sitkFloat64)
    )
    return filter_slice


def get_curr_otsu_slice() -> sitk.Image:
    """Return Otsu filtered slice of the current image.

    :return: Otsu filtered 2D rotated slice
    :rtype: sitk.Image"""
    rotated_slice: sitk.Image = get_curr_rotated_slice()
    filter_slice: sitk.Image = global_vars.OTSU_THRESHOLD_FILTER.Execute(
        sitk.Cast(rotated_slice, sitk.sitkFloat64)
    )
    return filter_slice


def get_curr_metadata() -> dict[str, str]:
    """Computes and returns currently displayed image's metadata.

    Note: Does not return all metadata stored in the file, just the metadata stored in sitk.Image.GetMetaDataKeys()
    For example, it's possible to get a sitk.Image's spacing using GetSpacing(), but the same spacing values won't
    be returned by this function.

    :return: metadata
    :rtype: dict[str, str]"""
    curr_img: sitk.Image = get_curr_image()
    rv: dict[str, str] = dict()
    for key in curr_img.GetMetaDataKeys():
        rv[key] = curr_img.GetMetaData(key)
    return rv


# TODO works only for NIFTI, not NRRD
def get_curr_physical_units() -> Union[str, None]:
    """Return currently displayed image's physical units from sitk.GetMetaData if it exists, else None.

    TODO works only for NIFTI, not NRRD.

    :return: units or None
    :rtype: str or None"""
    curr_img: sitk.Image = get_curr_image()
    if constants.NIFTI_METADATA_UNITS_KEY in curr_img.GetMetaDataKeys():
        return constants.NIFTI_METADATA_UNITS_VALUE_TO_PHYSICAL_UNITS[
            curr_img.GetMetaData(constants.NIFTI_METADATA_UNITS_KEY)
        ]
    return None


def get_curr_properties_tuple() -> tuple:
    """Return properties tuple for the currently loaded batch of images.

    :return: current properties tuple
    :rtype: tuple"""
    return get_properties_from_sitk_image(list(global_vars.IMAGE_DICT.values())[0])


def del_curr_img() -> None:
    """Remove currently displayed image from IMAGE_DICT.

    Decrements CURR_IMAGE_INDEX if removing the last element.

    Will not check for IMAGE_DICT being empty after the deletion (GUI should be disabled).
    This happens in the GUI.

    :return: None
    :rtype: None"""
    if len(global_vars.IMAGE_DICT) == 0:
        print("Can't remove from empty list!")
        return

    del global_vars.IMAGE_DICT[get_curr_path()]

    # Just deleted the last image. Index must decrease by 1
    if global_vars.CURR_IMAGE_INDEX == len(global_vars.IMAGE_DICT):
        global_vars.CURR_IMAGE_INDEX -= 1


def next_img() -> None:
    """Increment CURR_IMAGE_INDEX, wrapping if necessary.

    :return: None
    :rtype: None"""
    global_vars.CURR_IMAGE_INDEX = (global_vars.CURR_IMAGE_INDEX + 1) % len(
        global_vars.IMAGE_DICT
    )


def previous_img() -> None:
    """Decrement CURR_IMAGE_INDEX, wrapping if necessary.

    :return: None
    :rtype: None"""
    global_vars.CURR_IMAGE_INDEX = (global_vars.CURR_IMAGE_INDEX - 1) % len(
        global_vars.IMAGE_DICT
    )


def get_rotated_slice_hardcoded(
    mri_img_3d: sitk.Image,
    theta_x: int = 0,
    theta_y: int = 0,
    theta_z: int = 0,
    slice_num: int = 0,
) -> sitk.Image:
    """Get 2D rotated slice of mri_img_3d from hardcoded values. Rotation values are in degrees, slice_num is an int.

    For unit testing. Sets center of global EULER_3D_TRANSFORM to center of rotation of mri_img_3d.

    :param mri_img_3d:
    :type mri_img_3d: sitk.Image
    :param theta_x:
    :type theta_x: int
    :param theta_y:
    :type theta_y: int
    :param theta_z:
    :type theta_z: int
    :param slice_num:
    :type slice_num: int
    :return: 2D rotated slice using hardcoded settings
    :rtype: sitk.Image"""
    global_vars.EULER_3D_TRANSFORM.SetCenter(get_center_of_rotation(mri_img_3d))
    global_vars.EULER_3D_TRANSFORM.SetRotation(
        degrees_to_radians(theta_x),
        degrees_to_radians(theta_y),
        degrees_to_radians(theta_z),
    )
    return sitk.Resample(mri_img_3d, global_vars.EULER_3D_TRANSFORM)[:, :, slice_num]



"""Helper functions for image processing. Main algorithm."""

import SimpleITK as sitk
import cv2
import numpy as np

import NeuroRuler.utils.exceptions as exceptions
from NeuroRuler.utils.constants import (
    NUM_CONTOURS_IN_INVALID_SLICE,
    ThresholdFilter,
    BinaryColor,
)
import NeuroRuler.utils.gui_settings as settings
from NeuroRuler.utils.global_vars import (
    SMOOTHING_FILTER,
    OTSU_THRESHOLD_FILTER,
    BINARY_THRESHOLD_FILTER,
)

NUM_PIXELS_TO_CHECK_ON_EACH_EDGE_FOR_BACKGROUND_COLOR_DETECTION: int = 25
"""Roughly this many pixels are checked along the top, bottom, left, and right edges of the image."""
MAX_NUM_MISMATCHED_PIXELS_FOR_BACKGROUND_COLOR_DETECTION: int = 3
"""At most this many edge pixels can be different from the pixel at (0, 0)."""


# The RV is a np array, not sitk.Image
# because we can't actually use a sitk.Image contour in the rest of the process
# To compute arc length, we need a np array
# To overlay the contour on top of the base image in the GUI, we need a np array
def contour(
    img_2d: sitk.Image, threshold_filter: ThresholdFilter = ThresholdFilter.Otsu
) -> np.ndarray:
    r"""Generate the contour of a 2D slice by applying smoothing, Otsu threshold or binary threshold,
    hole filling, and island removal (select largest component). Return a binary (0|1) numpy
    array with only the points within the contour=1.

    Calls sitk.GetArrayFromImage() at the end, which will return the transpose of the sitk.Image.
    Consider whether to re-transpose the result or not.

    :param img_2d:
    :type img_2d: sitk.Image
    :param threshold_filter: ThresholdFilter.Otsu or ThresholdFilter.Binary. Defaults to ThresholdFilter.Otsu
    :type threshold_filter: ThresholdFilter
    :return: binary (0|1) numpy array with only the points on the contour = 1
    :rtype: np.ndarray"""
    smooth_slice: sitk.Image = SMOOTHING_FILTER.Execute(
        sitk.Cast(img_2d, sitk.sitkFloat64)
    )

    if threshold_filter == ThresholdFilter.Otsu:
        # This always results in fg = 0 (black), bg = 1 (white)
        thresholded: sitk.Image = OTSU_THRESHOLD_FILTER.Execute(smooth_slice)
    else:
        # This sometimes results in fg = 0 (black), bg = 1 (white)
        # other times fg = 1 (white), bg = 0 (black)
        # Depends on the lower and upper threshold settings
        thresholded: sitk.Image = BINARY_THRESHOLD_FILTER.Execute(smooth_slice)
        if (
            background_color_of_binary_thresholded_slice(thresholded)
            == BinaryColor.Black
        ):
            thresholded = sitk.NotImageFilter().Execute(thresholded)

    # Image needs to be inverted here (i.e., brain 0 black and background 1 white)
    # for BinaryGrindPeakImageFilter to work
    hole_filling: sitk.Image = sitk.BinaryGrindPeakImageFilter().Execute(thresholded)

    # BinaryGrindPeakImageFilter results in inverted foreground/background 0 and 1, need to invert
    inverted: sitk.Image = sitk.NotImageFilter().Execute(hole_filling)

    largest_component: sitk.Image = select_largest_component(inverted)

    contour: sitk.Image = sitk.BinaryContourImageFilter().Execute(largest_component)

    # GetArrayFromImage returns the transpose of the sitk representation
    return sitk.GetArrayFromImage(contour)


# Credit: https://discourse.itk.org/t/simpleitk-extract-largest-connected-component-from-binary-image/4958
def select_largest_component(binary_slice: sitk.Image) -> sitk.Image:
    r"""Remove islands from a binary (0|1) 2D slice. Return a binary slice
    containing only the largest connected component.

    :param binary_slice: Binary (0|1) 2D slice
    :type binary_slice: sitk.Image
    :return: Binary (0|1) slice with only the largest connected component
    :rtype: sitk.Image"""
    component_image: sitk.Image = sitk.ConnectedComponent(binary_slice)
    sorted_component_image: sitk.Image = sitk.RelabelComponent(
        component_image, sortByObjectSize=True
    )
    largest_component_binary_image: sitk.Image = sorted_component_image == 1
    return largest_component_binary_image


def length_of_contour(
    binary_contour_slice: np.ndarray, raise_exception: bool = True
) -> float:
    r"""Given a 2D binary slice, return the arc length of the parent contour.

    ``cv2.findContours`` will find all contours if there is more than one. Most valid brain slices have 2 or 3.

    The binary slice passed into this function should be processed by ``contour()``
    (i.e., hole-filling, island removal, threshold, etc.)
    to contain just one contour (except in edge cases that we don't need to worry about)
    to guarantee an accurate result.

    This function assumes the contour is a closed curve.

    :param binary_contour_slice: 2D binary (0|1) slice that should be pre-processed by contour() to contain just a single contour.
    :type binary_contour_slice: np.ndarray
    :param raise_exception: Whether or not to raise ComputeCircumferenceOfInvalidSlice.  Defaults to True and should be False only for unit testing purposes.
    :type raise_exception: bool
    :raise exceptions.ComputeCircumferenceOfInvalidSlice: If >= constants.NUM_CONTOURS_IN_INVALID_SLICE contours are detected.
    :return: Arc length of parent contour
    :rtype: float"""

    # contours is an array of contours
    # A single contour, contours[0], looks like [[[122  76]] [[121  77]] [[107  77]] ... [[106  78]]],
    # representing boundary points of the contour
    # contours[1] would be another contour, and so on
    # hierarchy is a list of the same length as contours that provides information about each contour
    # See the documentation for more detail
    contours, hierarchy = cv2.findContours(
        binary_contour_slice, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
    )
    num_contours: int = len(contours)

    if settings.DEBUG:
        print(
            f"Number of contours detected after processing: {num_contours} (in imgproc.length_of_contour())"
        )

    if raise_exception and num_contours >= NUM_CONTOURS_IN_INVALID_SLICE:
        raise exceptions.ComputeCircumferenceOfInvalidSlice(len(contours))

    # NOTE: select_largest_component removes all "islands" from the image.
    # But there can still be contours within the largest contour.
    # Most valid brain slices have 2 contours, rarely 3.
    # Assuming there are no islands, contours[0] is always the parent contour.
    # See unit test in test_imgproc.py: test_contours_0_is_always_parent_contour_if_no_islands
    parent_contour: np.ndarray = contours[0]
    # True means we assume the contour is a closed curve.
    arc_length = cv2.arcLength(parent_contour, True)
    return arc_length


def length_of_contour_with_spacing(
    binary_contour_slice: np.ndarray, x_spacing: float, y_spacing: float
) -> float:
    r"""Given a 2D binary slice (i.e., RV of contour()), return arc length of parent contour,
    accounting for x_spacing and y_spacing values.
    MAKE SURE x_spacing and y_spacing values are passed in the correct order!
    The binary slice passed into this function should be processed by ``contour()``
    to guarantee an accurate result.
    Slices for which circumference is calculated are always parallel to z,
    so the z spacing value is always ignored.
    This function assumes the contour is a closed curve.

    :param binary_contour_slice:
    :type binary_contour_slice: np.ndarray
    :param x_spacing:
    :type x_spacing: float
    :param y_spacing:
    :type y_spacing: float
    :raise: exceptions.ComputeCircumferenceOfInvalidSlice if contours detected >= constants.NUM_CONTOURS_IN_INVALID_SLICE
    :return: arc length of parent contour
    :rtype: float
    """
    contours, hierarchy = cv2.findContours(
        # TC89_L1 works better than SIMPLE
        # That is, when (a, a, a) pixel spacing image is converted to (a, b, c) pixel spacing via Slicer,
        # The arc length is closer to the (a, a, a) arc length if using TC89_L1 than when using SIMPLE
        binary_contour_slice,
        cv2.RETR_TREE,
        cv2.CHAIN_APPROX_TC89_L1,
    )

    num_contours: int = len(contours)
    if settings.DEBUG:
        print(
            f"Number of contours detected after processing: {num_contours} (in imgproc.length_of_contour())"
        )

    if num_contours >= NUM_CONTOURS_IN_INVALID_SLICE:
        raise exceptions.ComputeCircumferenceOfInvalidSlice(num_contours)

    parent_contour: np.ndarray = contours[0]

    arc_length: float = 0

    for i in range(len(parent_contour) - 1):
        # contours[0] would return [[x y]], so need to do an additional [0] index to just get [x y]
        arc_length += distance_2d_with_spacing(
            parent_contour[i][0], parent_contour[i + 1][0], x_spacing, y_spacing
        )
    # Get distance between first and last points
    arc_length += distance_2d_with_spacing(
        parent_contour[-1][0], parent_contour[0][0], x_spacing, y_spacing
    )
    return arc_length


def distance_2d_with_spacing(p1, p2, x_spacing: float, y_spacing: float) -> float:
    r"""Return the distance between two 2D iterables (list or tuple) given x and y spacing values.
    :param p1: 2D point
    :type p1: iterable
    :param p2: 2D point
    :type p2: iterable
    :param x_spacing:
    :type x_spacing: float
    :param y_spacing:
    :type y_spacing: float"""
    assert len(p1) == 2 and len(p2) == 2
    return np.sqrt(
        (x_spacing * (p1[0] - p2[0])) ** 2 + (y_spacing * (p1[1] - p2[1])) ** 2
    )


# TODO: Super naive, but realistically will work fine
def background_color_of_binary_thresholded_slice(img_2d: sitk.Image) -> BinaryColor:
    r"""Checks the background color of the slice returning from binary threshold filter
    since some settings result in fg 0, bg 1, and other settings result in fg 1, bg 0.

    Internally, compares the pixel value at (0, 0) to pixel values at the top, bottom, left, and right edges.
    Very naive approach that assumes there won't be noise or islands at the edges.

    :param img_2d:
    :type img_2d: sitk.Image
    :return: BinaryColor.Black or BinaryColor.White
    :rtype: BinaryColor"""
    width: int = img_2d.GetWidth()
    height: int = img_2d.GetHeight()
    background_color: int = img_2d.GetPixel(0, 0)
    mismatched_pixels: int = 0
    for i in range(
        1,
        width,
        width // NUM_PIXELS_TO_CHECK_ON_EACH_EDGE_FOR_BACKGROUND_COLOR_DETECTION,
    ):
        mismatched_pixels += img_2d.GetPixel(i, 0) != background_color
        mismatched_pixels += img_2d.GetPixel(i, height - 1) != background_color
    for j in range(
        1,
        height,
        height // NUM_PIXELS_TO_CHECK_ON_EACH_EDGE_FOR_BACKGROUND_COLOR_DETECTION,
    ):
        mismatched_pixels += img_2d.GetPixel(0, j) != background_color
        mismatched_pixels += img_2d.GetPixel(width - 1, j) != background_color
    if mismatched_pixels > MAX_NUM_MISMATCHED_PIXELS_FOR_BACKGROUND_COLOR_DETECTION:
        raise Exception(
            f"Could not successfully detect background color after executing binary threshold.\nMore than {MAX_NUM_MISMATCHED_PIXELS_FOR_BACKGROUND_COLOR_DETECTION} pixels on the top, bottom, left, or right edge don't match the pixel at the top left corner."
        )

    return BinaryColor(background_color)



"""Test stuff in imgproc.py.

A lot of tests here test behavior of libraries instead of our code. This was useful when learning how
to use the libraries, not so useful anymore.

We use == for float comparison instead of |a-b|<epsilon when the numbers should be *exactly* the same."""

import SimpleITK as sitk
import numpy as np
import cv2
import pytest
from pathlib import Path
from NeuroRuler.utils.imgproc import contour, length_of_contour
import NeuroRuler.utils.exceptions as exceptions
from NeuroRuler.utils.constants import (
    DATA_DIR,
    SUPPORTED_IMAGE_EXTENSIONS,
    degrees_to_radians,
)
from NeuroRuler.utils.global_vars import READER
from NeuroRuler.utils.img_helpers import (
    get_rotated_slice_hardcoded,
    get_center_of_rotation,
)

EPSILON: float = 0.001
"""Used for `float` comparisons."""

EXAMPLE_IMAGES: dict[Path, sitk.Image] = dict()
for extension in SUPPORTED_IMAGE_EXTENSIONS:
    for path in DATA_DIR.glob(extension):
        READER.SetFileName(str(path))
        EXAMPLE_IMAGES[path] = READER.Execute()


@pytest.mark.skip(reason="Doesn't need to run again unless new images are added")
def test_all_images_min_value_0_max_value_less_than_1600():
    for img in EXAMPLE_IMAGES.values():
        img_np: np.ndarray = sitk.GetArrayFromImage(img)
        assert img_np.min() == 0 and img_np.max() < 1600


@pytest.mark.skip(reason="Doesn't need to run again")
def test_dimensions_of_np_array_same_as_original_image_but_transposed():
    """Probably not needed but just in case.

    The dimensions of the numpy array are the same as the original image.

    Additionally, PNG files generated from numpy arrays (no metadata) look the same as slices of the original image (i.e., spacing correct).

    Pretty sure that means the arc length generated from the numpy array is the arc length of the original image, with the same units as the original image.
    """
    for img in EXAMPLE_IMAGES.values():
        for slice_z in range(img.GetSize()[2] // 5):
            slice = img[:, :, slice_z]
            # Transposed
            np_slice = sitk.GetArrayFromImage(slice)

            assert slice.GetSize()[0] == np_slice.shape[1]
            assert slice.GetSize()[1] == np_slice.shape[0]


@pytest.mark.skip(reason="Doesn't need to run again")
def test_numpy_2D_slice_array_is_transpose_of_sitk_2D_slice_array():
    """Confirm that the numpy matrix representation of a 2D slice is the transpose of the sitk matrix representation of a slice.

    Can ignore this test later."""
    for img in EXAMPLE_IMAGES.values():
        for z_slice in range(img.GetSize()[2] // 5):
            slice_sitk: sitk.Image = img[:, :, z_slice]
            slice_np: np.ndarray = sitk.GetArrayFromImage(slice_sitk)

            for i in range(slice_np.shape[0]):
                for j in range(slice_np.shape[1]):
                    assert slice_np[i][j] == slice_sitk.GetPixel(j, i)


@pytest.mark.skip(reason="Doesn't need to run for a while")
def test_contour_doesnt_mutate_slice():
    """Test that contour() doesn't mutate its argument."""
    for img in EXAMPLE_IMAGES.values():
        for slice_num in range(img.GetSize()[2] // 3):
            rotated_slice: sitk.Image = get_rotated_slice_hardcoded(
                img, 0, 0, 0, slice_num
            )
            rotated_slice_copy: sitk.Image = get_rotated_slice_hardcoded(
                img, 0, 0, 0, slice_num
            )
            contour(rotated_slice)
            for i in range(rotated_slice.GetSize()[0]):
                for j in range(rotated_slice.GetSize()[1]):
                    assert rotated_slice.GetPixel(i, j) == rotated_slice_copy.GetPixel(
                        i, j
                    )


@pytest.mark.skip(reason="Doesn't need to run again")
def test_contour_returns_binary_slice():
    """Test that the contour function always returns a binary (0|1) slice."""
    for img in EXAMPLE_IMAGES.values():
        for slice_num in range(img.GetSize()[2] // 5):
            rotated_slice = get_rotated_slice_hardcoded(img, 0, 0, 0, slice_num)
            contour_slice_np: np.ndarray = contour(rotated_slice)
            assert contour_slice_np.min() <= 1 and contour_slice_np.max() <= 1


@pytest.mark.skip(reason="Doesn't need to run again")
def test_contour_retranspose_has_same_dimensions_as_original_image():
    for img in EXAMPLE_IMAGES.values():
        for theta_x in range(0, 30, 15):
            for theta_y in range(0, 30, 15):
                for theta_z in range(0, 30, 15):
                    for slice_num in range(img.GetSize()[2] // 3):
                        rotated_slice = get_rotated_slice_hardcoded(
                            img, theta_x, theta_y, theta_z, slice_num
                        )
                        contour_slice: np.ndarray = contour(rotated_slice)
                        assert (
                            contour_slice.shape[0] == img.GetSize()[0]
                            and contour_slice.shape[1] == img.GetSize()[1]
                        )


@pytest.mark.skip(reason="This should be run again later")
def test_length_of_contour_doesnt_mutate_contour():
    for img in EXAMPLE_IMAGES.values():
        for slice_num in range(img.GetSize()[2] // 10):
            rotated_slice: sitk.Image = get_rotated_slice_hardcoded(
                img, 0, 0, 0, slice_num
            )
            contour_slice: np.ndarray = contour(rotated_slice)
            contour_slice_copy: np.ndarray = contour_slice.copy()
            length_of_contour(contour_slice, False)
            assert np.array_equal(contour_slice, contour_slice_copy)


@pytest.mark.skip(reason="Doesn't need to run again")
def test_contours_0_is_always_parent_contour_if_no_islands():
    """Assuming there are no islands in the image, then contours[0] results in the parent contour.

    See documentation on our wiki page about hierarchy. tl;dr hierarchy[0][i] returns information about the i'th contour.
    hierarchy[0][i][3] is information about the parent contour of the i'th contour. So if hierarchy[0][0][3] = -1, then the 0'th contour is the parent.
    """
    for img in EXAMPLE_IMAGES.values():
        for slice_num in range(img.GetSize()[2] // 7):
            rotated_slice: sitk.Image = get_rotated_slice_hardcoded(
                img, 0, 0, 0, slice_num
            )
            # contour removes islands
            contour_slice: np.ndarray = contour(rotated_slice)
            contours, hierarchy = cv2.findContours(
                contour_slice, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
            )
            assert hierarchy[0][0][3] == -1


@pytest.mark.skip(reason="Doesn't need to run again")
def test_arc_length_of_copy_after_transpose_same_as_no_copy_after_transpose():
    """Test arc length of two re-transposed arrays is the same when calling .copy() on one but not the other."""
    for img in EXAMPLE_IMAGES.values():
        for theta_x in range(0, 30, 15):
            for theta_y in range(0, 30, 15):
                for theta_z in range(0, 30, 15):
                    for slice_num in range(0, img.GetSize()[2], img.GetSize()[2] // 10):
                        rotated_slice: sitk.Image = get_rotated_slice_hardcoded(
                            img, theta_x, theta_y, theta_z, slice_num
                        )
                        contour_slice_retransposed_not_copied = contour(rotated_slice)
                        # The below duplicates work but it's to be safe
                        contour_slice_retransposed_copied = contour(
                            rotated_slice
                        ).copy()

                        # Assumes it's a closed curve but it might not be
                        length_of_not_copied = length_of_contour(
                            contour_slice_retransposed_not_copied, False
                        )
                        length_of_copied = length_of_contour(
                            contour_slice_retransposed_copied, False
                        )
                        assert length_of_not_copied == length_of_copied


@pytest.mark.skip(
    reason="User can see in the GUI the contour generated to confirm its accuracy. Also, this won't matter except for edge cases where the slice is invalid"
)
def test_arc_length_of_transposed_matrix_is_same_except_for_invalid_slice():
    """Per discussion here https://github.com/NIRALUser/NeuroRuler/commit/a230a6b57dc34ec433e311d760cc53841ddd6a49,

    Test that the arc length of a contour and its transpose is the same in a specific case. It probably generalizes to the general case.

    Specifically, for a matrix and its transpose, cv2.findContours will return [ [[x0 y0]] [[x1 y1]] [[x2 y2]] ... ] and [ [[y0 x0]], [[y1 x1]] [[y2 x2]] ... ]

    But cv2.arcLength will apply the distance formula to these contours and that will return the same result.

    However, if pixel spacing is off (non-square pixels), then the distance formula would need a scaling factor for one of the dimensions. Then we'd have to account for this.

    But the pixel spacing of the underlying `np.ndarray` passed into cv2.findContours *seems* to be fine. See discussion in the GH link.

    TODO: Unit test with pre-computed circumferences to really confirm this."""
    for img in EXAMPLE_IMAGES.values():
        # f.write(f"{DATA_DIR.name}/{img.path.name}\n")
        for theta_x in range(0, 31, 15):
            for theta_y in range(0, 31, 15):
                for theta_z in range(0, 31, 15):
                    for slice_num in range(0, img.GetSize()[2]):
                        rotated_slice: sitk.Image = get_rotated_slice_hardcoded(
                            img, theta_x, theta_y, theta_z, slice_num
                        )
                        contour_slice: np.ndarray = contour(rotated_slice)
                        # .copy() probably isn't needed if above test passes
                        contour_slice_transposed: np.ndarray = np.transpose(
                            contour_slice
                        )
                        try:
                            length_1 = length_of_contour(contour_slice)
                            length_2 = length_of_contour(contour_slice_transposed)
                            assert length_1 == length_2
                        except exceptions.ComputeCircumferenceOfInvalidSlice:
                            pass


@pytest.mark.skip(reason="")
def test_contour_slice_retranspose_same_dimensions_as_original_slice():
    """Test that the np array generated after contouring has the same dimensions as the original
    sitk slice.

    The process:

    1. MRI image
    2. SITK image
    3. np array

    Our concern is unit conversions between 1 & 2 and 2 & 3.

    Regarding 2, if the dimensions are the same and the aspect ratio looks correct in the GUI,
    does this mean the arc length of the np array is the same as that of the sitk slice
    without need for unit conversion? That is, the arc length of the np array is that of the sitk slice.

    Therefore, we need not worry about unit conversion between steps 2 and 3?

    Regarding 1, given that the GUI displays an image with correct aspect ratio, is the arc length of the
    sitk image the same as that of the physical brain?
    """
    for img in EXAMPLE_IMAGES.values():
        original_dimensions: tuple = img.GetSize()
        for theta_x in range(0, 31, 15):
            for theta_y in range(0, 31, 15):
                for theta_z in range(0, 31, 15):
                    for slice_num in range(
                        0, original_dimensions[2], original_dimensions[2] // 4
                    ):
                        rotated_slice: sitk.Image = get_rotated_slice_hardcoded(
                            img, theta_x, theta_y, theta_z, slice_num
                        )
                        binary_contour = contour(rotated_slice)
                        assert (
                            original_dimensions[0] == binary_contour.shape[0]
                            and original_dimensions[1] == binary_contour.shape[1]
                        )


@pytest.mark.skip(reason="Passed locally, doesn't need to run again")
def test_rotation_doesnt_affect_spacing():
    img = list(EXAMPLE_IMAGES.values())[0]
    e3d = sitk.Euler3DTransform()
    e3d.SetCenter(get_center_of_rotation(img))
    spacing = img.GetSpacing()
    for theta_x in range(0, 100, 25):
        for theta_y in range(0, 100, 25):
            for theta_z in range(0, 100, 25):
                e3d.SetRotation(
                    degrees_to_radians(theta_x),
                    degrees_to_radians(theta_y),
                    degrees_to_radians(theta_z),
                )
                new_img = sitk.Resample(img, e3d)
                assert new_img.GetSpacing() == spacing



"""GUI settings set through command-line arguments and ``gui_config.json``.

Command-line arguments override the values in the JSON.

There shouldn't be any blank (empty string) values here
(like the blank values in ``gui_config.json`` that are set to default values by ``parser.py``)
because if parsing the GUI json fails (e.g., doesn't exist),
then we need an actual working value here."""

from pathlib import Path
import NeuroRuler.utils.constants as constants

DEBUG: bool = False
"""Whether or not to print debugging information throughout execution."""

FILE_BROWSER_START_DIR: Path = Path.cwd()
"""The starting directory that is opened in the file browser.

Defaults to cwd."""

EXPORTED_FILE_NAMES_USE_INDEX: bool = False
"""If True, then exported files will be named using the index in the program.

E.g. 1_0_0_0_0.png, 2_90_180_0_60.csv, etc., where the first part is the file name and
the other parts are image settings.

If False (default), then exported files will be named using the file name of the original file.

E.g. MicroBiome_1month_T1w_0_0_0_0.png, MicroBiome_1month_T1w_90_180_0_60.csv."""

THEME_NAME: str = "dark-nr"
"""Name of theme in NeuroRuler/GUI/themes.

The full path to the .qss file is {constants.THEME_DIR}/{THEME_NAME}/stylesheet.qss."""

CONTOUR_COLOR: str = "b55162"
"""Color of the contour.

This can be a 6-hexit string rrggbb (don't prepend 0x) or a name (e.g. red, blue, etc.).

Internally, this is converted to a QColor using imgproc.string_to_QColor().
QColor supports 8-hexit rrggbbaa but doesn't work in our GUI, i.e. aa=00 appears fully bright in the GUI."""

STARTUP_WIDTH_RATIO: float = 0.65
"""Max GUI width as fraction of primary monitor width. Configurable in JSON"""
STARTUP_HEIGHT_RATIO: float = 0.6
"""Max GUI height as fraction of primary monitor height. Configurable in JSON"""

DISPLAY_ADVANCED_MENU_MESSAGES_IN_TERMINAL: bool = False
"""Whether the advanced menu messages display in terminal or new window"""



"""Constants used in tests."""

import os
import numpy as np

TARGET_R_SQUARED: float = 0.98
UBUNTU_GITHUB_ACTIONS_CI: bool = os.getenv("HOME") == "/home/runner"
"""Whether the tests are being run in an Ubuntu GitHub Actions CI environment.
Use this to skip PyQt6 imports if in Ubuntu GH Actions CI environment.

Ubuntu environments lack a required ``libEGL.so`` file for PyQt6 imports. This causes an error when importing PyQt6.
However, Windows environments have the required dependencies for importing PyQt6.

See GH actions 559 https://github.com/NIRALUser/NeuroRuler/actions/runs/4901894070/jobs/8753416763.
Note test_algorithm was skipped on Ubuntu but passed on Windows.

Therefore, this environment variable checks specifically for whether we're in an Ubuntu GH CI environment
(Windows paths use backslashes)."""


def labeled_result(path) -> float:
    with open(path, "r") as file:
        line = file.readline().strip()
        parts = line.split("\t")
        last_section = parts[-1]
        return float(last_section)


# Source: https://www.askpython.com/python/coefficient-of-determination
def compute_r_squared(actual: list[float], predicted: list[float]) -> float:
    return (np.corrcoef(actual, predicted)[0, 1]) ** 2



"""Defines ``main()``, the entrypoint of the CLI.

See ``tests/test_cli.py`` for examples of how to use this script.

Run with the ``-h`` option to see all CLI options."""

import NeuroRuler.utils.constants as constants
import NeuroRuler.utils.global_vars as global_vars
import NeuroRuler.utils.imgproc as imgproc
import NeuroRuler.utils.cli_settings as cli_settings
import SimpleITK as sitk
import numpy as np
from pathlib import Path
from typing import Union

from NeuroRuler.utils.img_helpers import (
    initialize_globals,
    get_curr_image,
    get_curr_rotated_slice,
    get_curr_physical_units,
)


def main() -> None:
    """Main entrypoint of CLI."""
    file_path = Path(cli_settings.FILE)

    initialize_globals([file_path])

    # slice options
    global_vars.THETA_X = cli_settings.THETA_X
    global_vars.THETA_Y = cli_settings.THETA_Y
    global_vars.THETA_Z = cli_settings.THETA_Z
    if (
        cli_settings.SLICE != -1
    ):  # initialize_globals will init it to middle slice if -1
        global_vars.SLICE = cli_settings.SLICE

    # smoothing options
    global_vars.SMOOTHING_FILTER.SetConductanceParameter(
        cli_settings.CONDUCTANCE_PARAMETER
    )
    global_vars.SMOOTHING_FILTER.SetNumberOfIterations(
        cli_settings.SMOOTHING_ITERATIONS
    )
    global_vars.SMOOTHING_FILTER.SetTimeStep(cli_settings.TIME_STEP)

    rotated_slice: sitk.Image = get_curr_rotated_slice()

    # Binary threshold setting application
    # Won't be used if cli_settings.THRESHOLD_FILTER is Otsu, but applied anyway
    global_vars.BINARY_THRESHOLD_FILTER.SetLowerThreshold(
        cli_settings.LOWER_BINARY_THRESHOLD
    )
    global_vars.BINARY_THRESHOLD_FILTER.SetUpperThreshold(
        cli_settings.UPPER_BINARY_THRESHOLD
    )

    binary_contour_slice: np.ndarray = imgproc.contour(
        rotated_slice, cli_settings.THRESHOLD_FILTER
    )

    spacing: tuple = get_curr_image().GetSpacing()
    units: Union[str, None] = get_curr_physical_units()
    circumference: float = imgproc.length_of_contour_with_spacing(
        binary_contour_slice, spacing[0], spacing[1]
    )

    if cli_settings.RAW:
        print(circumference)
    else:
        print(
            f"Calculated Circumference: {round(circumference, constants.NUM_DIGITS_TO_ROUND_TO)} {units if units is not None else constants.MESSAGE_TO_SHOW_IF_UNITS_NOT_FOUND}"
        )


if __name__ == "__main__":
    import NeuroRuler.utils.parser as parser

    parser.parse_cli_config()
    parser.parse_cli()
    main()



"""Constant values and functions. DO NOT MUTATE ANY VARIABLE IN THIS FILE FROM OUTSIDE OF THIS FILE!

This file should not import any module in this repo to avoid circular imports."""

from pathlib import Path
import warnings
import functools
from screeninfo import get_monitors, ScreenInfoError
from numpy import pi
from typing import Union
from enum import Enum
import pkg_resources

OUTPUT_DIR: Path = Path("output")
"""Directory for storing output."""
if not OUTPUT_DIR.exists():
    OUTPUT_DIR.mkdir()

JSON_CLI_CONFIG_PATH: Path = Path("cli_config.json")
"""Settings that configure cli_settings.py."""
if not JSON_CLI_CONFIG_PATH.exists():
    # __name__ will get to the utils module
    # and cli_config.json is at root directory
    JSON_CLI_CONFIG_PATH = Path(
        pkg_resources.resource_filename(__name__, "../../cli_config.json")
    )

JSON_GUI_CONFIG_PATH: Path = Path("gui_config.json")
"""Settings that configure gui_settings.py."""
if not JSON_GUI_CONFIG_PATH.exists():
    # __name__ will get to the utils module
    # and gui_config.json is at root directory
    JSON_GUI_CONFIG_PATH = Path(
        pkg_resources.resource_filename(__name__, "../../gui_config.json")
    )

DATA_DIR: Path = Path("data")

SUPPORTED_IMAGE_EXTENSIONS: tuple = (".nii.gz", ".nii", ".nrrd")
"""Image file formats supported. Must be a subset of the file formats supported by SimpleITK."""

THEME_DIR: Path = Path("NeuroRuler") / "GUI" / "themes"
"""themes/ directory where .qss stylesheets and resources.py files are stored."""
if not THEME_DIR.exists():
    try:
        THEME_DIR = Path(pkg_resources.resource_filename("NeuroRuler.GUI", "themes"))
    # This will occur on circular import, which will occur when running
    # from NeuroRuler.CLI import CLI
    # but CLI doesn't care about THEME_DIR
    # This is a janky way to avoid circular import
    except ImportError:
        pass

THEMES: list[str] = []
"""List of themes, i.e. the names of the directories in THEME_DIR."""

# Without this, autodocumentation might crash
# THEME_DIR obviously exists at this point, except maybe in autodocumentation code
if THEME_DIR.exists():
    for path in THEME_DIR.iterdir():
        if path.is_dir():
            THEMES.append(path.name)
    THEMES = sorted(THEMES)
else:
    pass


class View(Enum):
    """X, Y, or Z view.

    The letters are assigned indices 0-2 to be used for indexing operations."""

    X = 0
    Y = 1
    Z = 2


class ThresholdFilter(Enum):
    """Determines the threshold filter (Otsu or binary) used in imgproc.contour()."""

    Otsu = 0
    Binary = 1


class BinaryColor(Enum):
    """Self-explanatory"""

    Black = 0
    White = 1


ROTATION_MIN: int = -90
"""In degrees"""
ROTATION_MAX: int = 90
"""In degrees"""


NUM_CONTOURS_IN_INVALID_SLICE: int = 10
"""If this number of contours or more is detected in a slice after processing by contour()
(Otsu, largest component, etc.), then the slice is considered invalid."""

NIFTI_METADATA_UNITS_VALUE_TO_PHYSICAL_UNITS: dict[str, str] = {
    "0": "unknown",
    "1": "meters (m)",
    "2": "millimeters (mm)",
    "3": "microns (μm)",
    "8": "seconds (s)",
    "16": "milliseconds (ms)",
    "24": "microseconds (μs)",
    "32": "Hertz (Hz)",
    "40": "parts-per-million (ppm)",
    "48": "radians per second (rad/s)",
}
"""Maps the value of ``xyzt_units`` of the metadata of a NIfTI file to physical meaning.

Based on https://brainder.org/2012/09/23/the-nifti-file-format/.

See img_helpers.py MRIImage for code to get metadata using sitk."""

NIFTI_METADATA_UNITS_KEY: str = "xyzt_units"
"""In the NIfTI metadata dictionary, the numerical str value attached to this key represents units of the file."""

NUM_DIGITS_TO_ROUND_TO: int = 3
"""For floats, number of digits to round to, i.e. round(float, n)."""


# Got these values by looking at ITK-SNAP and
# https://simpleitk.org/doxygen/latest/html/classitk_1_1simple_1_1DICOMOrientImageFilter.html#details
# Two characters in each string orients the image properly (i.e., same as ITK-SNAP)
# The remaining character affects direction of rotations (CCW vs CW)

# TODO: X rotation goes the wrong way :(, but it's also wrong for LPI
# Last 2 are the important ones
X_ORIENTATION_STR: str = "RPI"
"""Orientation string to pass into sitk.DICOMOrientImageFilter to orient X view correctly."""
# First and third are the important ones
Y_ORIENTATION_STR: str = "LPI"
"""Orientation string to pass into sitk.DICOMOrientImageFilter to orient Y view correctly."""
# First 2 are the important ones
Z_ORIENTATION_STR: str = "LPS"
"""Orientation string to pass into sitk.DICOMOrientImageFilter to orient Z view correctly."""
ORIENTATION_STRINGS: tuple[str, str, str] = (
    X_ORIENTATION_STR,
    Y_ORIENTATION_STR,
    Z_ORIENTATION_STR,
)
"""(X_ORIENTATION_STR, Y_ORIENTATION_STR, Z_ORIENTATION_STR)

Intended to be indexed using View.X.value, View.Y.value, and View.Z.value."""

PRIMARY_MONITOR_DIMENSIONS: tuple[int, int] = (500, 500)
"""Set to user's primary monitor's dimensions. 500, 500 are dummy values"""

try:
    for monitor in get_monitors():
        if monitor.is_primary:
            PRIMARY_MONITOR_DIMENSIONS = (monitor.width, monitor.height)
            break
except ScreenInfoError:
    # This will occur in GH automated tests.
    pass


MESSAGE_TO_SHOW_IF_UNITS_NOT_FOUND: str = "millimeters (mm)"
"""We assume units are millimeters if we can't find units in metadata"""


# Source: https://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically
def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)  # turn off filter
        warnings.warn(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


def degrees_to_radians(angle: Union[int, float]) -> float:
    """It's quite simple.

    :param num: A degree measure
    :type num: int or float
    :return: Equivalent radian measure
    :rtype: float"""
    return angle * pi / 180



"""Black box integration tests for the command line.

Hardcoded numbers were manually retrieved from the GUI calculations.

The tests in ``tests/imports_GUI`` confirm that the GUI calculations are correct. So if the CLI calculations
match those of the GUI, then the CLI calculations are correct."""

import subprocess
from subprocess import PIPE


def test_basic():
    command = "python cli.py data/IBIS_Case1_V06_t1w_RAI.nrrd"
    proc = subprocess.Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
    assert (
        proc.stdout.read().rstrip()
        == b"Calculated Circumference: 433.436 millimeters (mm)"
    )


def test_slice_options():
    command = (
        "python cli.py --x=16 --y=2 --z=22 --slice=96 data/IBIS_Case1_V06_t1w_RAI.nrrd"
    )
    proc = subprocess.Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
    assert (
        proc.stdout.read().rstrip()
        == b"Calculated Circumference: 421.511 millimeters (mm)"
    )


def test_binary():
    command = "python cli.py --slice=69 --lower=0.0 --upper=200.0 --filter=binary data/BCP_Dataset_2month_T1w.nrrd"
    proc = subprocess.Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
    assert (
        proc.stdout.read().rstrip()
        == b"Calculated Circumference: 390.444 millimeters (mm)"
    )


def test_explicit_otsu():
    """Otsu is applied by default, but this tests for the explicit flag."""
    command = "python cli.py --filter=otsu data/IBIS_Dataset_NotAligned_6month_T1w.nrrd"
    proc = subprocess.Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
    assert (
        proc.stdout.read().rstrip()
        == b"Calculated Circumference: 425.13 millimeters (mm)"
    )


def test_smoothing_options():
    command = "python cli.py --conductance=1.0 --iterations=20 --step=0.05 data/MicroBiome_1month_T1w.nii.gz"
    proc = subprocess.Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
    assert (
        proc.stdout.read().rstrip()
        == b"Calculated Circumference: 365.712 millimeters (mm)"
    )



# Source: https://github.com/hmeine/qimage2ndarray/blob/master/setup.py

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

install_requires: list[str] = [
    "setuptools",
    "SimpleITK",
    "numpy",
    "argparse",
    "opencv-python",
    "pytest",
    "PyQt6",
    "qimage2ndarray",
    "screeninfo",
]
"""All required (i.e., for functionality) dependencies that are installed when running `pip install NeuroRuler`.

Non-functional (e.g., formatting, documentation) dependencies listed in requirements.txt."""

setup(
    name="NeuroRuler",
    # For current version number, see
    # https://pypi.org/project/NeuroRuler/
    version="1.0.0",
    description="A program that calculates head circumference from MRI data (`.nii`, `.nii.gz`, `.nrrd`).",
    # Cannot use multiple authors
    # https://stackoverflow.com/questions/9999829/how-to-specify-multiple-authors-emails-in-setup-py
    author="COMP523 Team D",
    author_email="comp523d@gmail.com",
    url="https://github.com/NIRALUser/NeuroRuler",
    download_url="https://github.com/NIRALUser/NeuroRuler/releases",
    keywords=[
        "MRI",
        "NIfTI",
        "NRRD",
        "brain",
        "circumference",
        "PyQt6",
    ],
    install_requires=install_requires,
    tests_require=install_requires + ["tox", "pytest", "pytest-cov"],
    # See https://setuptools.pypa.io/en/latest/userguide/package_discovery.html
    package_dir={"NeuroRuler": "NeuroRuler"},
    packages=find_packages(),
    package_data={
        "NeuroRuler": ["GUI/*.ui", "GUI/static/*", "GUI/themes/*/*", "../*.json"]
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



"""Test algorithm by computing R^2 from GUI calculation and ground truth data.

Simulates GUI button clicks.

Uses GUI. GUI imports and tests will not run in CI. See note in tests/README.md."""

from tests.constants import (
    TARGET_R_SQUARED,
    labeled_result,
    UBUNTU_GITHUB_ACTIONS_CI,
    compute_r_squared,
)
import re
import os
import sys
import os
import pytest
from pathlib import Path
from typing import Union
import SimpleITK as sitk
import numpy as np

if not UBUNTU_GITHUB_ACTIONS_CI:
    from PyQt6.QtTest import QTest
    from PyQt6.QtWidgets import (
        QPushButton,
        QApplication,
    )
    from PyQt6.uic.load_ui import loadUi
    from PyQt6.QtCore import Qt, QSize
    from NeuroRuler.utils.img_helpers import *
    from NeuroRuler.GUI.main import *
    import NeuroRuler.utils.global_vars as global_vars
    from NeuroRuler.utils.constants import View, ThresholdFilter
    import NeuroRuler.utils.constants as constants

    # Note, do not use imports like
    # from NeuroRuler.utils.global_vars import IMAGE_DICT
    # This would make the global variables not work
    import NeuroRuler.utils.global_vars as global_vars
    import NeuroRuler.utils.imgproc as imgproc
    import NeuroRuler.utils.gui_settings as gui_settings


@pytest.mark.skipif(
    UBUNTU_GITHUB_ACTIONS_CI, reason="No GUI on Ubuntu GitHub Actions CI environment"
)
def calculate_circumference(path) -> float:
    # Set up UI
    app = QApplication(sys.argv)
    window = MainWindow()

    # Load in image
    window.browse_files(False, path)
    window.set_view_z()
    window.orient_curr_image()
    window.toggle_setting_to_false()
    gui_settings.CONTOUR_COLOR = "b55162"
    binary_contour_slice: np.ndarray = window.render_curr_slice()
    return window.render_circumference(binary_contour_slice)


@pytest.mark.skipif(
    UBUNTU_GITHUB_ACTIONS_CI, reason="No GUI on Ubuntu GitHub Actions CI environment"
)
def test_algorithm():
    labeled_data = []
    calculated_data = []
    data_folder = "data"

    length = 0

    for file_name in os.listdir(data_folder):
        if file_name.endswith("w.nrrd"):
            nrrd_prefix = os.path.splitext(file_name)[0][
                :10
            ]  # Remove the file extension to get the prefix
            for tsv_file_name in os.listdir(data_folder):
                if nrrd_prefix in tsv_file_name and tsv_file_name.endswith(".tsv"):
                    length = length + 1
                    labeled_data.append(
                        labeled_result(data_folder + "/" + tsv_file_name)
                    )
                    calculated_data.append(
                        calculate_circumference(data_folder + "/" + file_name)
                    )
                    break
            else:
                continue

    assert length > 20
    assert compute_r_squared(labeled_data, calculated_data) > TARGET_R_SQUARED



"""After installing NeuroRuler via pip, the functions here are importable like so

from NeuroRuler import {function}

where NeuroRuler is the name of the package this __init__.py file is in."""

import sys
import os
import NeuroRuler.GUI.main as main
import NeuroRuler.utils.parser as parser


def gui() -> None:
    """Start GUI."""
    # Source: https://stackoverflow.com/questions/5047734/in-osx-change-application-name-from-python
    if sys.platform.startswith("darwin"):
        # Set app name, if PyObjC is installed
        # Python 2 has PyObjC preinstalled
        # Python 3: pip3 install pyobjc-framework-Cocoa
        try:
            from Foundation import NSBundle

            bundle = NSBundle.mainBundle()
            if bundle:
                app_name = os.path.splitext(os.path.basename(sys.argv[0]))[0]
                app_info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
                if app_info:
                    app_info["CFBundleName"] = app_name
        except ImportError:
            pass

    # Source: https://stackoverflow.com/questions/1551605/how-to-set-applications-taskbar-icon-in-windows-7/1552105#1552105
    # myappid: Use unicode
    import ctypes

    myappid = "mycompany.myproduct.subproduct.version"  # arbitrary string
    try:
        # For Windows.
        # App icon works without this on macOS.
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    except:
        pass

    parser.parse_gui_config()
    parser.parse_gui_cli()
    main.main()



"""After installing NeuroRuler via pip, the functions here are importable like so

from NeuroRuler.CLI import {function}

where NeuroRuler.CLI is the name of the package this __init__.py file is in."""

import NeuroRuler.CLI.main as main
import NeuroRuler.utils.parser as parser


def cli() -> None:
    """Run CLI."""
    parser.parse_cli_config()
    parser.parse_cli()
    main.main()
